from NuRadioReco.modules.base.module import register_run
import numpy as np
from scipy import constants, stats

from radiotools import helper as hp

from NuRadioReco.utilities import units, ice
from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.parameters import electricFieldParameters as efp

import logging
logger = logging.getLogger('efieldTimeDirectionFitter')


class efieldTimeDirectionFitter:
    """
    Calculates basic signal parameters.
    """

    def __init__(self):
        self.begin()
        pass

    def begin(self, debug=False, time_uncertainty=0.1 * units.ns):
        self.__debug = debug
        self.__time_uncertainty = time_uncertainty
        pass

    @register_run()
    def run(self, evt, station, det, debug=True, channels_to_use=[0, 1, 2, 3], cosmic_ray=False):
        station_id = station.get_id()

        times = []
        times_error = []
        positions = []
        for iCh, efield in enumerate(station.get_electric_fields()):
            if(len(efield.get_channel_ids()) > 1):
                # FIXME: this can be changed later if each efield has a position and absolute time
                raise AttributeError("found efield that is valid for more than one channel. Position can't be determined.")
            channel_id = efield.get_channel_ids()[0]
            if(channel_id not in channels_to_use):
                continue
            times.append(efield[efp.signal_time])
            if(efield.has_parameter_error(efp.signal_time)):
                times_error.append((efield.get_parameter_error(efp.signal_time) ** 2 + self.__time_uncertainty ** 2) ** 0.5)
            else:
                times_error.append(self.__time_uncertainty)
            positions.append(det.get_relative_position(station_id, channel_id))

        times = np.array(times)
        times_error = np.array(times_error)
        positions = np.array(positions)
        site = det.get_site(station_id)
        n_ice = ice.get_refractive_index(-0.01, site)

        from scipy import optimize as opt

        def get_expected_times(params, positions):
            zenith, azimuth = params
            if cosmic_ray:
                if((zenith < 0) or (zenith > 0.5 * np.pi)):
                    return np.ones(len(positions)) * np.inf
            else:
                if((zenith < 0.5 * np.pi) or (zenith > np.pi)):
                    return np.ones(len(positions)) * np.inf
            v = hp.spherical_to_cartesian(zenith, azimuth)
            c = constants.c * units.m / units.s
            if not cosmic_ray:
                c = c / n_ice
                logger.debug("using speed of light = {:.4g}".format(c))
            t_expected = -(np.dot(v, positions.T) / c)
            return t_expected

        def obj_plane(params, positions, t_measured):
            t_expected = get_expected_times(params, positions)
            chi2 = np.sum(((t_expected - t_expected.mean()) - (t_measured - t_measured.mean())) ** 2 / times_error ** 2)
            logger.debug("texp = {texp}, tm = {tmeas}, {chi2}".format(texp=t_expected, tmeas=t_measured, chi2=chi2))
            return chi2

        method = "Nelder-Mead"
        options = {'maxiter': 1000,
                   'disp': False}
        zenith_start = 135 * units.deg
        if cosmic_ray:
            zenith_start = 45 * units.deg
        starting_chi2 = {}
        for starting_az in np.array([0, 90, 180, 270]) * units.degree:
            starting_chi2[starting_az] = obj_plane((zenith_start, starting_az), positions, times)
        azimuth_start = min(starting_chi2, key=starting_chi2.get)
        res = opt.minimize(obj_plane, x0=[zenith_start, azimuth_start], args=(positions, times), method=method, options=options)

        chi2 = res.fun
        df = len(channels_to_use) - 3
        if(df == 0):
            chi2ndf = chi2
            chi2prob = stats.chi2.sf(chi2, 1)
        else:
            chi2ndf = chi2 / df
            chi2prob = stats.chi2.sf(chi2, df)

        output_str = "reconstucted angles theta = {:.1f}, phi = {:.1f}, chi2/ndf = {:.2g}/{:d} = {:.2g}, chi2prob = {:.3g}".format(res.x[0] / units.deg,
                                                                                              hp.get_normalized_angle(res.x[1]) / units.deg,
                                                                                              res.fun, df,
                                                                                              chi2ndf,
                                                                                              chi2prob)

        logger.info(output_str)
        station[stnp.zenith] = res.x[0]
        station[stnp.azimuth] = hp.get_normalized_angle(res.x[1])
        station[stnp.chi2_efield_time_direction_fit] = chi2
        station[stnp.ndf_efield_time_direction_fit] = df
        if(cosmic_ray):
            station[stnp.cr_zenith] = res.x[0]
            station[stnp.cr_azimuth] = hp.get_normalized_angle(res.x[1])

        if(self.__debug):
            # calculate residuals
            t_exp = get_expected_times(res.x, positions)
            from matplotlib import pyplot as plt
            fig, ax = plt.subplots(1, 1)
            ax.errorbar(channels_to_use, ((times - times.mean()) - (t_exp - t_exp.mean())) / units.ns, fmt='o',
                        yerr=times_error / units.ns)
            ax.set_xlabel("channel id")
            ax.set_ylabel(r"$t_\mathrm{meas} - t_\mathrm{exp}$ [ns]")
            pass

    def end(self):
        pass



import NuRadioReco.modules.io.araroot.readARAData
from NuRadioReco.modules.base.module import setup_logger
import logging
import matplotlib.pyplot as plt
from NuRadioReco.utilities import units

logger = setup_logger()

readARAData = NuRadioReco.modules.io.araroot.readARAData.readARAData()

n=readARAData.begin("/home/uzair/Documents/AraRootStuff/event9129.root")
print(n)

for iE,evt in enumerate(readARAData.run()):
    print(evt.get_id())
    channel = evt.get_station(2).get_channel(0)
    plt.figure()
    plt.plot(channel.get_times()/units.ns,channel.get_trace()/units.mV)
    plt.show()
    break



from __future__ import absolute_import, division, print_function  # , unicode_literals
from dash.dependencies import Input, Output, State
import dash_core_components as dcc
import dash_html_components as html
import dash
import plotly.graph_objs as go
import json
import numpy as np
import uuid
import glob
from NuRadioReco.utilities import units
from flask import Flask, send_from_directory
import os
import sys
from NuRadioReco.detector import detector
# from apps import summary
import logging
import logging
from NuRadioReco.modules.base import module
logger = module.setup_logger(level=logging.INFO)

det = detector.Detector(source='sql')

server = Flask(__name__)
app = dash.Dash(__name__, server=server)

# Loading screen CSS
# app.css.append_css({"external_url": "https://codepen.io/chriddyp/pen/brPBPO.css"})
app.css.append_css({
    'external_url': 'https://codepen.io/chriddyp/pen/bWLwgP.css'
})
# app.css.append_css({"external_url": "https://maxcdn.bootstrapcdn.com/font-awesome/4.7.0/css/font-awesome.min.css"})


app.title = 'ARIANNA detector database'

app.layout = html.Div([
    # represents the URL bar, doesn't render anything
    dcc.Location(id='url', refresh=False),


    dcc.Dropdown(id='station_id',
             options=[{'label': l, 'value': l} for l in [14, 15, 17, 18, 19, 30, 32, 50, 51, 52, 61]]),

    html.Div(id='main')
])

keys = ['channel_id', 'commission_time', 'decommission_time',
        'ant_position_x', 'ant_position_y', 'ant_position_z',
        'ant_type',
        'ant_comment',
        'ant_orientation_phi', 'ant_orientation_theta',
        'ant_rotation_phi', 'ant_rotation_theta',
        'ant_deployment_time',
        'amp_reference_measurement', 'amp_type',
        'cab_id', 'cab_length', 'cab_reference_measurement', 'cab_time_delay', 'cab_type',
        'adc_id', 'adc_n_samples', 'adc_nbits', 'adc_sampling_frequency', 'adc_time_delay',
        ]

# next/previous buttons
@app.callback(
Output('main', 'children'),
[Input('station_id', 'value')])
def show_station(station_id):
    if not det.has_station(station_id):
        return html.H2(children="station {} not present in data base".format(station_id))

    tts = det.get_unique_time_periods(station_id)
    d = []
    d.append(html.H1(children="station {}".format(station_id)))
    for iT, t0 in enumerate(tts[:-1]):
        t1 = tts[iT+1]
        d.append(html.H4(children="station configuration from {} - {}".format(t0, t1)))
        det.update(t0)
        c = det.get_channel(station_id, 0)
        header = [html.Tr([html.Th(col) for col in keys])]

        for iC in det.get_channel_ids(station_id):
            c = det.get_channel(station_id, iC)
            header.append(html.Tr([html.Td(c[key]) for key in keys]))
        d.append(html.Table(header))

    res =html.Div(d)
    return res

if __name__ == '__main__':
    app.run_server(debug=True, port=8080)



import time
import copy
from scipy import signal, fftpack
import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize as opt
import logging

from radiotools import helper as hp
from radiotools import plthelpers as php

import NuRadioReco.framework.base_trace
from NuRadioReco.utilities import trace_utilities
from NuRadioReco.utilities import ice
from NuRadioReco.detector import antennapattern

from NuRadioReco.framework import electric_field as ef
from NuRadioReco.utilities import geometryUtilities as geo_utl
from NuRadioReco.utilities import units
from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.parameters import electricFieldParameters as efp
import NuRadioReco.modules.voltageToEfieldConverterPerChannel
import NuRadioReco.modules.electricFieldBandPassFilter



electricFieldBandPassFilter = NuRadioReco.modules.electricFieldBandPassFilter.electricFieldBandPassFilter()
voltageToEfieldConverterPerChannel = NuRadioReco.modules.voltageToEfieldConverterPerChannel.voltageToEfieldConverterPerChannel()
voltageToEfieldConverterPerChannel.begin()

def get_array_of_channels(station, det, zenith, azimuth, polarization):
    """
    Returns an array of the channel traces that is cut to the physical overlapping time

    Parameters
    ----------
    zenith: float
        Arrival zenith angle at antenna
    azimuth: float
        Arrival azimuth angle at antenna
    polarization: int
        0: eTheta
        1: ePhi
    """
    time_shifts = np.zeros(8)
    t_geos = np.zeros(8)

    sampling_rate = station.get_channel(0).get_sampling_rate()
    station_id = station.get_id()
    site = det.get_site(station_id)
    for iCh, channel in enumerate(station.get_electric_fields()):
        channel_id = channel.get_channel_ids()[0]
        antenna_position = det.get_relative_position(station_id, channel_id)
        # determine refractive index of signal propagation speed between antennas
        refractive_index = ice.get_refractive_index(1, site)  # if signal comes from above, in-air propagation speed
        if(zenith > 0.5 * np.pi):
            refractive_index = ice.get_refractive_index(antenna_position[2], site)  # if signal comes from below, use refractivity at antenna position
        refractive_index = 1.353
        time_shift = -geo_utl.get_time_delay_from_direction(zenith, azimuth, antenna_position, n=refractive_index)
        t_geos[iCh] = time_shift
        time_shift += channel.get_trace_start_time()
        time_shifts[iCh] = time_shift

    delta_t = time_shifts.max() - time_shifts.min()
    tmin = time_shifts.min()
    tmax = time_shifts.max()
    trace_length = station.get_electric_fields()[0].get_times()[-1] - station.get_electric_fields()[0].get_times()[0]

    traces = []
    n_samples = None
    for iCh, channel in enumerate(station.get_electric_fields()):
        tstart = delta_t - (time_shifts[iCh] - tmin)
        tstop = tmax - time_shifts[iCh] - delta_t + trace_length
        iStart = int(round(tstart * sampling_rate))
        iStop = int(round(tstop * sampling_rate))
        if(n_samples is None):
            n_samples = iStop - iStart
            if(n_samples % 2):
                n_samples -= 1

        trace = copy.copy(channel.get_trace()[polarization])  # copy to not modify data structure
        trace = trace[iStart:(iStart + n_samples)]
        base_trace = NuRadioReco.framework.base_trace.BaseTrace()  # create base trace class to do the fft with correct normalization etc.
        base_trace.set_trace(trace, sampling_rate)
        traces.append(base_trace)

    return traces

class beamFormingDirectionFitter:
    """
    Fits the direction using interferometry between desired channels.
    """

    def __init__(self):
        self.__zenith = []
        self.__azimuth = []
        self.__delta_zenith = []
        self.__delta_azimuth = []
        self.begin()
        self.logger = logging.getLogger("NuRadioReco.beamFormingDirectionFitter")

    def begin(self, debug=False, log_level=None):
        if(log_level is not None):
            self.logger.setLevel(log_level)
        self.__debug = debug

    def run(self, evt, station, det, polarization, n_index=None,channels=[4,5,6,7], ZenLim=[90 * units.deg, 180 * units.deg],
            AziLim=[0 * units.deg, 360 * units.deg]):
        """
        reconstruct signal arrival direction for all events through beam forming.
        https://arxiv.org/pdf/1009.0345.pdf

        Parameters
        ----------
        polarization: int
            0: eTheta
            1: ePhi
        n_index: float
            the index of refraction
        channels: list
            the channel ids
        ZenLim: 2-dim array/list of floats
            the zenith angle limits for the fit
            default if 0-90deg (upward coming signal)
        AziLim: 2-dim array/list of floats
            the azimuth angle limits for the fit
            default is 0-360deg
        """

        def ll_regular_station(angles, evt, station, det, polarization, sampling_rate, positions, channels):
            """
            Likelihood function for a four antenna ARIANNA station, using correction.
            Using correlation, has no built in wrap around, pulse needs to be in the middle
            """

            zenith = angles[0]
            azimuth = angles[1]

            station.set_parameter(stnp.zenith,zenith)
            station.set_parameter(stnp.azimuth,azimuth)
            station.set_electric_fields([]) # resets EFields, necessary
            voltageToEfieldConverterPerChannel.run(evt, station, det,pol=polarization,debug=False)  # Antenna response
            electricFieldBandPassFilter.run(evt, station, det,passband=[120 * units.MHz, 300 * units.MHz], filter_type='butterabs')

            Efields_object = get_array_of_channels(station, det, zenith, azimuth, polarization+1)
            Efields = []
            Efield_Times = []
            maximum = 0 # location at maximum among all traces
            for chan in channels:
                efield = Efields_object[chan]
                Efield_Times.append(efield.get_times())
                Efield = efield.get_trace()
                if max(Efield) > maximum:
                    maximum = max(Efield)
                Efields.append(Efield)#np.pad(Efield, (200,200), 'constant', constant_values=(0, 0)))

            Efields[0] = Efields[0]/maximum # normalize all traces to remove small antenna responses, see last line of next for loop
            for i in range(len(Efields)-1):
                Efields[i+1] = Efields[i+1]/maximum

            N = len(Efields)
            N_pairs = 0.5*np.math.factorial(N)/np.math.factorial(N-2)
            cc = np.zeros(len(Efields[0]))
            for i in range(N-1): # finds the cc-beam taken from referenced paper above
                for j in range(N-1-i):
                    cc = cc + Efields[i]*Efields[j+1+i]
            cc = cc/N_pairs
            cc = np.sign(cc)*np.sqrt(np.abs(cc))

            cc = np.abs(cc)
            ave_cc = np.zeros_like(cc)
            n_bins = 2000
            ave_cc = np.convolve(cc, np.ones((n_bins))/float(n_bins),mode='same')

            likelihood = -1*np.max(ave_cc)
            return likelihood


        station_id = station.get_id()
        positions_all = det.get_relative_positions(station_id)
        positions = []
        for chan in channels:
            positions.append(positions_all[chan])
        sampling_rate = station.get_channel(0).get_sampling_rate()

        ll = opt.brute(ll_regular_station, ranges=(slice(ZenLim[0], ZenLim[1], 1.0*units.deg),
                                                   slice(AziLim[0], AziLim[1], 1.0*units.deg)),
                        args=(evt, station, det, polarization, sampling_rate, positions, channels),
                        full_output=True, finish=opt.fmin)  # slow but does the trick

        station[stnp.zenith] = max(ZenLim[0], min(ZenLim[1], ll[0][0]))
        station[stnp.azimuth] = ll[0][1]

        if self.__debug:
            import peakutils

            # Show fit space
            zen = np.arange(ZenLim[0], ZenLim[1], 1*units.deg)
            az = np.arange(AziLim[0], AziLim[1], 1*units.deg)

            x_plot = np.zeros(zen.shape[0] * az.shape[0])
            y_plot = np.zeros(zen.shape[0] * az.shape[0])
            z_plot = np.zeros(zen.shape[0] * az.shape[0])
            i = 0
            for z in zen:
                for a in az:
                    # Evaluate fit function for grid
                    z_plot[i] = ll_regular_station([z, a], evt, station, det, polarization, sampling_rate, positions, channels)
                    x_plot[i] = a
                    y_plot[i] = z
                    i += 1

            fig, ax = plt.subplots(1, 1)
            ax.scatter(np.asarray(x_plot)/units.deg, np.asarray(y_plot)/units.deg, c=z_plot, cmap='gnuplot2_r', lw=0)
            ax.scatter(np.rad2deg(ll[0][1]), np.rad2deg(ll[0][0]), marker='o', label='Fit')
            ax.colorbar(label='Fit parameter')
            ax.set_ylabel('Zenith [rad]')
            ax.set_xlabel('Azimuth [rad]')
            plt.tight_layout()
            plt.legend()
            plt.show()

    def end(self):
        pass



from __future__ import absolute_import, division, print_function  # , unicode_literals
import time
import radiotools.helper as hp
import dash_html_components as html
import dataprovider
from NuRadioReco.utilities import units
from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.parameters import channelParameters as chp
from NuRadioReco.framework.parameters import electricFieldParameters as efp
from NuRadioReco.eventbrowser.apps.common import get_point_index
from NuRadioReco.eventbrowser.default_layout import default_layout
import NuRadioReco.eventbrowser.apps.overview_plots.template_correlation
import NuRadioReco.eventbrowser.apps.overview_plots.rec_directions
import NuRadioReco.eventbrowser.apps.overview_plots.station_properties
import NuRadioReco.eventbrowser.apps.overview_plots.channel_properties
import NuRadioReco.eventbrowser.apps.overview_plots.electric_field_properties
import NuRadioReco.eventbrowser.apps.overview_plots.trigger_properties
import NuRadioReco.eventbrowser.apps.overview_plots.event_overview
import numpy as np
import logging
logger = logging.getLogger('overview')

provider = dataprovider.DataProvider()

layout = html.Div([
    html.Div([
        html.Div('Event Overview', className='panel-heading'),
        html.Div(NuRadioReco.eventbrowser.apps.overview_plots.event_overview.layout,
            className='panel-body')
    ], className='panel panel-default'),
    html.Div([
        html.Div([
            html.Div('Station', className='panel-heading'),
            html.Div(NuRadioReco.eventbrowser.apps.overview_plots.station_properties.layout,
            className='panel-body')
        ], className='panel panel-default', style={'flex':'1'}),
        html.Div([
            html.Div('Channels', className='panel-heading'),
            html.Div(NuRadioReco.eventbrowser.apps.overview_plots.channel_properties.layout,
            className='panel-body')
        ], className='panel panel-default', style={'flex':'1'}),
        html.Div([
            html.Div('Electric Fields', className='panel-heading'),
            html.Div(NuRadioReco.eventbrowser.apps.overview_plots.electric_field_properties.layout,
            className='panel-body')
        ], className='panel panel-default', style={'flex':'1'}),
        html.Div([
            html.Div('Triggers', className='panel-heading'),
            html.Div(NuRadioReco.eventbrowser.apps.overview_plots.trigger_properties.layout,
            className='panel-body')
        ], className='panel panel-default', style={'flex': '1'})
    ], style={'display': 'flex', 'margin': '20px 0'}),
    html.Div([
        html.Div('Correlations', className='panel-heading'),
        html.Div(NuRadioReco.eventbrowser.apps.overview_plots.template_correlation.layout,
        className='panel-body')
    ], className = 'panel panel-default'),
    html.Div([
        html.Div('Reconstructed Directions', className='panel-heading'),
        html.Div(NuRadioReco.eventbrowser.apps.overview_plots.rec_directions.layout,
        className='panel-body')
    ], className='panel panel-default')
])



from NuRadioReco.detector.ARA import analog_components
from NuRadioReco.modules.base.module import register_run
import numpy as np
from NuRadioReco.utilities import units
import time
import logging

logger = logging.getLogger("hardwareResponseIncorporator")


class hardwareResponseIncorporator:
    """
    Incorporates the gain and phase induced by the ARA hardware.


    """

    def __init__(self):
        self.__debug = False
        self.__time_delays = {}
        self.__t = 0
        self.begin()

    def begin(self, debug=False):
        self.__debug = debug

    @register_run()
    def run(self, evt, station, det, sim_to_data=False):
        """
        Switch sim_to_data to go from simulation to data or otherwise.
        """
        t = time.time()
        station_id = station.get_id()
        channels = station.iter_channels()

        for channel in channels:

            frequencies = channel.get_frequencies()
            system_response = analog_components.get_system_response(frequencies)
            trace_fft = channel.get_frequency_spectrum()

            if sim_to_data:

                trace_after_system_fft = trace_fft * system_response['gain'] * system_response['phase']
                # zero first bins to avoid DC offset
                trace_after_system_fft[0] = 0
                channel.set_frequency_spectrum(trace_after_system_fft, channel.get_sampling_rate())

            else:
                trace_before_system_fft = np.zeros_like(trace_fft)
                trace_before_system_fft[np.abs(system_response['gain']) > 0] = trace_fft[np.abs(system_response['gain']) > 0] / (system_response['gain'] * system_response['phase'])[np.abs(system_response['gain']) > 0]
                channel.set_frequency_spectrum(trace_before_system_fft, channel.get_sampling_rate())

        self.__t += time.time() - t

    def end(self):
        from datetime import timedelta
        logger.setLevel(logging.INFO)
        dt = timedelta(seconds=self.__t)
        logger.info("total time used by this module is {}".format(dt))
        return dt




import dash
import json
import plotly
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output, State
from app import app
from NuRadioReco.utilities import units
from NuRadioReco.eventbrowser.default_layout import default_layout
import numpy as np
from NuRadioReco.framework.parameters import stationParameters as stnp
import dataprovider
provider = dataprovider.DataProvider()

layout = [
    html.Div([
        html.Div([
            dcc.RadioItems(
                id='xcorrelation-event-type',
                options=[
                    {'label': 'Neutrino', 'value': 'nu'},
                    {'label': 'Cosmic Ray', 'value': 'cr'}
                ],
                value='nu'
            )
        ], style={'flex': 'none', 'padding-right': '20px'}),
        html.Div([
            dcc.Dropdown(
                id='cr-xcorrelation-dropdown',
                options=[]
            )
        ], style={'flex': '1'})
    ], style={'display': 'flex'}),
    html.Div([
        html.Div([
            dcc.Graph(id='cr-xcorrelation'),
        ], style={'flex': '1'}),
        html.Div([
            dcc.Graph(id='cr-xcorrelation-amplitude'),
        ], style={'flex': '1'})
    ], style={'display': 'flex'})
]

cr_xcorr_options = [
    {'label': 'maximum cr x-corr all channels', 'value': 'cr_max_xcorr'},
    {'label': 'maximum of avg cr x-corr in parallel cr channels', 'value': 'cr_avg_xcorr_parallel_crchannels'},
    {'label': 'maximum cr x-corr cr channels', 'value': 'cr_max_xcorr_crchannels'},
    {'label': 'average cr x-corr cr channels', 'value': 'cr_avg_xcorr_crchannels'},
]
nu_xcorr_options = [
    {'label': 'maximum nu x-corr all channels', 'value': 'nu_max_xcorr'},
    {'label': 'maximum of avg nu x-corr in parallel nu channels', 'value': 'nu_avg_xcorr_parallel_nuchannels'},
    {'label': 'maximum nu x-corr nu channels', 'value': 'nu_max_xcorr_nuchannels'},
    {'label': 'average nu x-corr nu channels', 'value': 'nu_avg_xcorr_nuchannels'}
]


@app.callback(Output('cr-xcorrelation-dropdown', 'options'),
            [Input('xcorrelation-event-type', 'value')])
def set_xcorrelation_options(event_type):
    if event_type == 'nu':
        return nu_xcorr_options
    else:
        return cr_xcorr_options


@app.callback(Output('cr-xcorrelation', 'figure'),
              [Input('cr-xcorrelation-dropdown', 'value'),
               Input('filename', 'value'),
               Input('event-ids', 'children'),
               Input('station-id-dropdown', 'value'),
               Input('xcorrelation-event-type', 'value')],
              [State('user_id', 'children')])
def plot_corr(xcorr_type, filename, jcurrent_selection, station_id, event_type, juser_id):
    if filename is None or station_id is None or xcorr_type is None:
        return {}
    user_id = json.loads(juser_id)
    ariio = provider.get_arianna_io(user_id, filename)
    fig = plotly.subplots.make_subplots(rows=1, cols=1)
    keys = ariio.get_header()[station_id].keys()
    if event_type == 'nu':
        if not stnp.nu_xcorrelations in keys:
            return {}
        xcorrs = ariio.get_header()[station_id][stnp.nu_xcorrelations]
    else:
        if not stnp.cr_xcorrelations in keys:
            return {}
        xcorrs = ariio.get_header()[station_id][stnp.cr_xcorrelations]
    if stnp.station_time in keys:
        times = []
        for time in ariio.get_header()[station_id][stnp.station_time]:
            times.append(time.value)
        fig.append_trace(plotly.graph_objs.Scatter(
            x=times,
            y=[xcorrs[i][xcorr_type] for i in range(len(xcorrs))],
            text=[str(x) for x in ariio.get_event_ids()],
            customdata=[x for x in range(ariio.get_n_events())],
            mode='markers',
            opacity=1
        ),1,1)
    else:
        return {}
    current_selection = json.loads(jcurrent_selection)
    if current_selection != []:
        for trace in traces:
            trace['selectedpoints'] = current_selection
    fig['layout'].update(default_layout)
    fig['layout']['yaxis'].update({'title': xcorr_type, 'range': [0, 1]})
    fig['layout']['hovermode'] = 'closest'
    return fig

@app.callback(Output('cr-xcorrelation-amplitude', 'figure'),
              [Input('cr-xcorrelation-dropdown', 'value'),
               Input('filename', 'value'),
               Input('event-ids', 'children'),
               Input('xcorrelation-event-type', 'value'),
               Input('station-id-dropdown', 'value')],
              [State('user_id', 'children')])
def plot_corr_amplitude(xcorr_type, filename, jcurrent_selection, event_type, station_id, juser_id):
    if filename is None or station_id is None or xcorr_type is None:
        return {}
    user_id = json.loads(juser_id)
    ariio = provider.get_arianna_io(user_id, filename)
    fig = plotly.subplots.make_subplots(rows=1, cols=1)
    keys = ariio.get_header()[station_id].keys()
    if event_type == 'nu':
        if not stnp.nu_xcorrelations in keys:
            return {}
        xcorrs = ariio.get_header()[station_id][stnp.nu_xcorrelations]
    else:
        if not stnp.cr_xcorrelations in keys:
            return {}
        xcorrs = ariio.get_header()[station_id][stnp.cr_xcorrelations]
    if stnp.channels_max_amplitude in keys:
        fig.append_trace(plotly.graph_objs.Scatter(
            x=ariio.get_header()[station_id][stnp.channels_max_amplitude] / units.mV,
            y=[xcorrs[i][xcorr_type] for i in range(len(xcorrs))],
            text=[str(x) for x in ariio.get_event_ids()],
            customdata=[x for x in range(ariio.get_n_events())],
            mode='markers',
            opacity=1
        ),1,1)
    else:
        return {}
    # update with current selection
    current_selection = json.loads(jcurrent_selection)
    if current_selection != []:
        for trace in traces:
            trace['selectedpoints'] = current_selection
    fig['layout'].update(default_layout)
    fig['layout']['xaxis'].update({'type': 'log', 'title': 'maximum amplitude [mV]'})
    fig['layout']['yaxis'].update({'title': xcorr_type, 'range': [0, 1]})
    fig['layout']['hovermode'] = 'closest'
    return fig



from NuRadioReco.detector import antennapattern
from NuRadioReco.utilities import units

from matplotlib import pyplot as plt
import numpy as np

# name of the antenna model
# see ./antenna_models_hash.json for available antenna models
antenna_model = "createLPDA_100MHz_InfAir"

# ploting range
min_freq, max_freq = 0 * units.MHz, 1000 * units.MHz
df = 1 * units.MHz
ff = np.arange(min_freq, max_freq, df)

# signal income direction
zenith, azimuth = np.deg2rad(0), np.deg2rad(0)

# antenna orientation (0, 0, 90, 0 : Upwards looking, northwards pointing lpda)
zen_boresight, azi_boresight, zen_ori, azi_ori = np.deg2rad(0), np.deg2rad(0), np.deg2rad(90), np.deg2rad(0)

provider = antennapattern.AntennaPatternProvider()
antenna = provider.load_antenna_pattern(antenna_model)

VEL = antenna.get_antenna_response_vectorized(ff, zenith, azimuth, zen_boresight, azi_boresight, zen_ori, azi_ori)

fig, ax = plt.subplots(1, 1)
ax.plot(ff / units.MHz, np.abs(VEL['theta']), '.-', label='eTheta')
ax.plot(ff / units.MHz, np.abs(VEL['phi']), '.-', label='ePhi')
ax.set_xlabel("frequency / MHz")
ax.set_ylabel("vector effective length / m")
fig.tight_layout()
ax.legend()
plt.savefig("VEL_%s_%d_%d.png" % (antenna_model, np.deg2rad(zenith), np.deg2rad(azimuth)))



import numpy as np
from NuRadioReco.utilities import units
import NuRadioReco.utilities.diodeSimulator

def get_window_around_maximum(station,
                              diode=None,
                              triggered_channels=None,
                              ratio = 0.01,
                              edge=20*units.ns):
    """
    This function filters the signal using a diode model and calculates
    the times around the filtered maximum where the signal is the ratio
    parameter times the maximum. Then, it creates a time window by substracting
    and adding the edge parameter to these times. This function is useful
    for reducing the probability of noise-triggered events while using the
    envelope phased array, although it can also be applied to the standard
    phased array.

    Parameters
    ----------
    station: NuRadioReco station
        Station containing the information on the detector
    diode: diodeSimulator or None
        Diode model to be applied. If None, a diode with an output 200 MHz low-pass
        filter is chosen
    triggered_channels: array of integers
        Channels used for the trigger, and also for creating the window
    ratio: float
        Cut parameter
    edge: float
        Times to be sustracted from the points defined by the ratio cut to
        create the window

    Returns:
    (left_time, right_time): (float, float) tuple
        Tuple containing the edges of the time window
    """

    if diode == None:
        diode_passband = (None, 200*units.MHz)
        diode = NuRadioReco.utilities.diodeSimulator.diodeSimulator(diode_passband)

    left_times = []
    right_times = []

    if (triggered_channels == None):
    	triggered_channels = [channel._id for channel in station.iter_channels()]

    for channel in station.iter_channels():  # loop over all channels (i.e. antennas) of the station

        if channel.get_id() not in triggered_channels:
            continue

        times = channel.get_times()

        trace = diode.tunnel_diode(channel)
        trace_max = np.max(np.max(trace))
        argmax = np.argmax(trace)
        trace_min = np.min(np.min(trace))
        argmin = np.argmin(trace)

        if (argmin == len(trace)-1):
            argmin = len(trace)-2
        if (argmin == argmax):
            if (argmin + 5 < len(trace)):
                argmax = argmin + 5
            else:
                argmax = len(trace) - 1
        if (argmin == 0):
            argmin = 1

        left_bin = argmin - np.argmin(np.abs(trace[0:argmin][::-1]+ratio*trace_max))
        left_times.append(times[left_bin])

        right_bin = argmax + np.argmin(np.abs(trace[argmax:None]-ratio*trace_max))
        right_times.append(times[right_bin])

    left_time = np.min(left_times) - edge
    right_time = np.max(right_times) + edge

    return (left_time,right_time)



#!/usr/bin/env python
import json
import NuRadioReco.modules.io.eventReader
import argparse
import numpy as np
import sys

parser = argparse.ArgumentParser()
parser.add_argument('--create_reference', help='create new reference instead of comparing to current reference', action='store_true')
args = parser.parse_args()

event_reader = NuRadioReco.modules.io.eventReader.eventReader()
event_reader.begin('NuRadioReco/test/trigger_tests/trigger_test_output.nur')

trigger_names = ['default_simple_threshold', 'default_high_low', 'default_multi_high_low', 'simple_phased_threshold']
properties = ['triggered', 'trigger_time', 'triggered_channels']
trigger_results = {}

if not args.create_reference:
    with open('NuRadioReco/test/trigger_tests/reference.json', 'r') as f:
        reference = json.load(f)

for event in event_reader.run():
    station = event.get_station(1)
    for trigger_name in trigger_names:
        trigger = station.get_trigger(trigger_name)
        if trigger_name not in trigger_results.keys():
            trigger_results[trigger_name] = {}
        for property in properties:
            if property not in trigger_results[trigger_name].keys():
                trigger_results[trigger_name][property] = []
            trigger_results[trigger_name][property].append(trigger.get_trigger_settings()[property])

found_error = False
if args.create_reference:
    with open('NuRadioReco/test/trigger_tests/reference.json', 'w') as f:
        json.dump(trigger_results, f, sort_keys=True,
                  indent=4, separators=(',', ': '))
else:
    for trigger_name in trigger_names:
        for property in properties:
            if(property == "trigger_time"):
                try:
                    np.testing.assert_allclose(np.array(trigger_results[trigger_name][property], dtype=np.float), np.array(reference[trigger_name][property], dtype=np.float))
                except AssertionError as e:
                    print('Property {} of trigger {} differs from reference'.format(property, trigger_name))
                    print(e)
                    found_error = True
            else:
                try:
                    np.testing.assert_equal(trigger_results[trigger_name][property], reference[trigger_name][property])
                except AssertionError as e:
                    print('Property {} of trigger {} differs from reference'.format(property, trigger_name))
                    print(e)
                    found_error = True

if found_error:
    sys.exit(-1)
else:
    print('Trigger test passed without issues')



from NuRadioReco.detector.antennapattern import *
import matplotlib.pyplot as plt
import logging
import copy

from NuRadioReco.modules.base import module
logger = module.setup_logger(level=logging.DEBUG)

provider = AntennaPatternProvider()

theta = 90 * units.deg
ff = np.linspace(50 * units.MHz, 1 * units.GHz, 1000)
bicone = provider.load_antenna_pattern("bicone_v8_inf_n1.78")
bicone_n14 = provider.load_antenna_pattern("bicone_v8_inf_n1.4")
bicone_air = provider.load_antenna_pattern("bicone_v8_InfAir")

bicone_XFDTD = provider.load_antenna_pattern("XFDTD_Vpol_CrossFeed_150mmHole_n1.78")

VELs = bicone.get_antenna_response_vectorized(ff, 90 * units.deg, np.deg2rad(0),
                                              np.deg2rad(180), 0, np.deg2rad(90), np.deg2rad(0))
fig, (ax) = plt.subplots(1, 1, sharey=True)
ax.plot(ff /units.MHz, np.abs(VELs['theta']), label='eTheta bicone n=1.78 (WIPL-D)')
# ax.plot(ff / units.MHz, np.abs(VELs['phi']), label='ePhi bicone (WIPL-D)')

VELs = bicone_XFDTD.get_antenna_response_vectorized(ff, 90 * units.deg, np.deg2rad(0),
                                                  np.deg2rad(180), 0, np.deg2rad(90), np.deg2rad(0))
ax.plot(ff / units.MHz, np.abs(VELs['theta']), '--', label='eTheta bicone n=1.78 (XFDTD)')
# ax.plot(ff / units.MHz, np.abs(VELs['phi']), '--', label='ePhi bicone (old ARA)')

VELs = bicone_air.get_antenna_response_vectorized(ff, 90 * units.deg, np.deg2rad(0),
                                                  np.deg2rad(180), 0, np.deg2rad(90), np.deg2rad(0))
ax.plot(ff / units.MHz, np.abs(VELs['theta']), '--', label='eTheta bicone (air)')
ax.plot(ff / units.MHz, np.abs(VELs['phi']), '--', label='ePhi bicone (air)')

ax.set_title('NS ant down, signal from East at 90deg')
ax.legend()
ax.set_ylabel("Heff [m]")
ax.set_xlabel("frequency [MHz]")


plt.show()



#!/usr/bin/env python
import json
import NuRadioReco.modules.io.eventReader
import argparse
import numpy as np
import sys
from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.parameters import channelParameters as chp
from NuRadioReco.framework.parameters import electricFieldParameters as efp

parser = argparse.ArgumentParser()
parser.add_argument('filename', help='name of the file that should be compared to reference')
parser.add_argument('reference_file', help='name of the reference file')
parser.add_argument('--station_id', default=32)
parser.add_argument('--create_reference', help='create new reference instead of comparing to current reference', action='store_true')
args = parser.parse_args()

event_reader = NuRadioReco.modules.io.eventReader.eventReader()
event_reader.begin(args.filename)

parameter_values = {}

for event in event_reader.run():
    parameter_values[f"{event.get_id():d}"] = {
        'station_parameters': {},
        'sim_station_parameters': {},
        'channel_parameters': {},
        'electric_field_parameters': {}
    }
    station = event.get_station(args.station_id)
    sim_station = station.get_sim_station()
    for param_name in stnp:
        if station.has_parameter(param_name):
            parameter_values[f"{event.get_id():d}"]['station_parameters'][param_name.name] = station.get_parameter(param_name)
        else:
            parameter_values[f"{event.get_id():d}"]['station_parameters'][param_name.name] = None
        if sim_station.has_parameter(param_name):
            parameter_values[f"{event.get_id():d}"]['sim_station_parameters'][param_name.name] = sim_station.get_parameter(param_name)
        else:
            parameter_values[f"{event.get_id():d}"]['sim_station_parameters'][param_name.name] = None
    for channel in station.iter_channels():
        for param_name in chp:
            if param_name.name not in parameter_values[f"{event.get_id():d}"]['channel_parameters'].keys():
                parameter_values[f"{event.get_id():d}"]['channel_parameters'][param_name.name] = []
            if channel.has_parameter(param_name):
                parameter_values[f"{event.get_id():d}"]['channel_parameters'][param_name.name].append(channel.get_parameter(param_name))
            else:
                parameter_values[f"{event.get_id():d}"]['channel_parameters'][param_name.name].append(None)
    for electric_field in station.get_electric_fields():
        for param_name in efp:
            if param_name.name not in parameter_values[f"{event.get_id():d}"]['electric_field_parameters'].keys():
                parameter_values[f"{event.get_id():d}"]['electric_field_parameters'][param_name.name] = []
            if electric_field.has_parameter(param_name):
                value = electric_field.get_parameter(param_name)
                if isinstance(value, np.ndarray):
                    value = list(value)
                parameter_values[f"{event.get_id():d}"]['electric_field_parameters'][param_name.name].append(value)
            else:
                parameter_values[f"{event.get_id():d}"]['electric_field_parameters'][param_name.name].append(None)



def assertDeepAlmostEqual(expected, actual, *args, **kwargs):
    """
    Assert that two complex structures have almost equal contents.

    Compares lists, dicts and tuples recursively. Checks numeric values
    using test_case's :py:meth:`unittest.TestCase.assertAlmostEqual` and
    checks all other values with :py:meth:`unittest.TestCase.assertEqual`.
    Accepts additional positional and keyword arguments and pass those
    intact to assertAlmostEqual() (that's how you specify comparison
    precision).

    :param test_case: TestCase object on which we can call all of the basic
    'assert' methods.
    :type test_case: :py:class:`unittest.TestCase` object
    """
    is_root = not '__trace' in kwargs
    trace = kwargs.pop('__trace', 'ROOT')
    try:
        if isinstance(expected, (int, float, complex)):
            np.testing.assert_allclose(actual, expected, *args, **kwargs)
        elif isinstance(expected, (list, tuple, np.ndarray)):
            np.testing.assert_equal(len(expected), len(actual))
            for index in range(len(expected)):
                v1, v2 = expected[index], actual[index]
                assertDeepAlmostEqual(v1, v2,
                                      __trace=repr(index), *args, **kwargs)
        elif isinstance(expected, dict):
            for key in actual:
                if key not in expected:
                    print('Key {} found in reconstruction but not in reference. '\
                        'This is fine if you just added it, but you should update '\
                        'the reference file to include it in future tests!'.format(key))
                else:
                    assertDeepAlmostEqual(expected[key], actual[key],
                                      __trace=repr(key), *args, **kwargs)
        else:
            np.testing.assert_equal(actual, expected)
    except AssertionError as exc:
        exc.__dict__.setdefault('traces', []).append(trace)
        if is_root:
            trace = ' -> '.join(reversed(exc.traces))
            exc = AssertionError("%s\nTRACE: %s" % (exc, trace))
        raise exc


if args.create_reference:
    with open(args.reference_file, 'w') as f:
        json.dump(parameter_values, f)
else:
    with open(args.reference_file, 'r') as f:
        parameter_reference = json.load(f)
        found_error = False
        assertDeepAlmostEqual(parameter_reference, parameter_values, rtol=1e-6)



from NuRadioReco.modules.base.module import register_run
import logging
logger = logging.getLogger('eventTypeIdentifier')


class eventTypeIdentifier:
    """
    Use this module to distinguish cosmic ray events from neutrino events
    """

    def begin(self):
        pass

    @register_run()
    def run(self, event, station, mode, forced_event_type='neutrino'):
        """
        Determines the type of event so that the correct modules can be
        executed accordingly. Currently neutrino and cosmic ray events are
        supported, but others can be added in the future.

        Parameters
        ---------------------
        event: event

        station: station

        mode: string
            specifies which criteria the decision to flag the event as cosmic ray
            or neutrino should be based on. Currently only supports 'forced', 
            which automatically flags the event to a specified type. Add new 
            modes to use different criteria
        forced_event_type: string
            if the mode is set to forced, this is the type the event will be set to.
            Currently supported options are neutrino and cosmic_ray
        """
        if mode == 'forced':
            if forced_event_type == 'neutrino':
                station.set_is_neutrino()
            elif forced_event_type == 'cosmic_ray':
                station.set_is_cosmic_ray()
            else:
                ValueError('Unsupported event type {}'.format(forced_event_type))
        else:
            raise ValueError('Unsupported mode {}'.format(mode))




import dash
import json
import plotly
from NuRadioReco.utilities import units
from NuRadioReco.eventbrowser.default_layout import default_layout
import numpy as np
from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.parameters import electricFieldParameters as efp
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output, State
from app import app
import dataprovider
provider = dataprovider.DataProvider()

layout = [
    dcc.Graph(id='cr-skyplot'),
]

@app.callback(Output('cr-skyplot', 'figure'),
              [Input('filename', 'value'),
               Input('trigger', 'children'),
               Input('event-ids', 'children'),
               Input('btn-open-file', 'value'),
               Input('station-id-dropdown', 'value')],
              [State('user_id', 'children')])
def cosmic_ray_skyplot(filename, trigger, jcurrent_selection, btn, station_id, juser_id):
    if filename is None or station_id is None:
        return {}
    user_id = json.loads(juser_id)
    current_selection = json.loads(jcurrent_selection)
    ariio = provider.get_arianna_io(user_id, filename)
    traces = []
    keys = ariio.get_header()[station_id].keys()
    if stnp.cr_zenith in keys and stnp.cr_azimuth in keys:
        traces.append(plotly.graph_objs.Scatterpolar(
            r=np.rad2deg(ariio.get_header()[station_id][stnp.cr_zenith]),
            theta=np.rad2deg(ariio.get_header()[station_id][stnp.cr_azimuth]),
            text=[str(x) for x in ariio.get_event_ids()],
            mode='markers',
            name='cosmic ray events',
            opacity=1,
            customdata=[x for x in range(ariio.get_n_events())],
            marker=dict(
                color='blue'
            )
        ))
    if current_selection != []:
        for trace in traces:
            trace['selectedpoints'] = current_selection
    return {
        'data': traces,
        'layout': plotly.graph_objs.Layout(
            showlegend=True,
            hovermode='closest',
            height=500
        )
    }



from NuRadioReco.modules.base.module import register_run
import numpy as np
import fractions
from decimal import Decimal
from NuRadioReco.utilities import units
from scipy import signal
import matplotlib.pyplot as plt
from radiotools import helper as hp
from NuRadioReco.utilities import templates
from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.parameters import channelParameters as chp

import logging
logger = logging.getLogger('channelTemplateCorrelation')

import matplotlib.pyplot as plt


class channelTemplateCorrelation:
    """
    Calculates correlation of waveform with neutrino/cr templates
    """

    def __init__(self, template_directory):
        self.__max_upsampling_factor = 5000
        self.__templates = templates.Templates(template_directory)
        self.begin()

    def begin(self, debug=False):
        self.__cr_templates = {}
        self.__ref_cr_template = {}
        self.__debug = debug

    def match_sampling(self, ref_template, resampling_factor):
        if(resampling_factor.numerator != 1):
            ref_template_resampled = signal.resample(ref_template, resampling_factor.numerator * len(ref_template))
        else:
            ref_template_resampled = ref_template
        if(resampling_factor.denominator != 1):
            ref_template_resampled = signal.resample(ref_template_resampled, len(ref_template_resampled) / resampling_factor.denominator)
        return ref_template_resampled

    @register_run()
    def run(self, evt, station, det, channels_to_use=[0, 1, 2, 3], cosmic_ray=False,
            n_templates=1):
        """
        Parameters
        -----------
        cosmic_ray: bool
            Switch for cosmic ray and neutrino analysis. Default is neutrino templates.
        n_templates: int
            default is 1: use just on standard template for all channels
            if n_templates is larger than one multiple templates are used and the
            average cross correlation for all templates is calculated. The set
            of templates contains several coreas input pulses and different
            incoming directions. The index first loops first over 6 different
            coreas pulses with different frequency content
            and then over azimuth angles of 0, 22.5 and 45 degree
            and then over zenith angles of 60, 50 and 70 degree
        """
        station_id = station.get_id()
        run = evt.get_run_number()
        event_id = int(evt.get_id())

        if n_templates == 1:

            if cosmic_ray:
                ref_template = self.__templates.get_cr_ref_template(station_id)
                ref_str = 'cr'
            else:
                ref_template = self.__templates.get_nu_ref_template(station_id)
                ref_str = 'nu'
        else:
            logger.debug("Using average of correlation over many templates")
            if cosmic_ray:
                ref_templates = self.__templates.get_set_of_cr_templates_full(station_id, n=n_templates)
                ref_str = 'cr'
            else:
                ref_templates = self.__templates.get_set_of_nu_templates_full(station_id, n=n_templates)
                ref_str = 'nu'

        xcorrs = []
        xcorrs_max = []

        for iCh, channel in enumerate(station.iter_channels()):
            channel_id = channel.get_id()
            xcorrs_ch = []
            xcorrpos_ch = []
            xcorrelations = {}

            orig_binning = 1. / det.get_sampling_frequency(station_id, channel_id)
            target_binning = 1. / channel.get_sampling_rate()
            resampling_factor = fractions.Fraction(Decimal(orig_binning / target_binning)).limit_denominator(self.__max_upsampling_factor)

            times = channel.get_times()
            trace = channel.get_trace()
            dt = times[1] - times[0]

            if n_templates == 1:

                ref_template_resampled = self.match_sampling(ref_template, resampling_factor)
                xcorr_trace = hp.get_normalized_xcorr(trace, ref_template_resampled)

                xcorrpos = np.argmax(np.abs(xcorr_trace))
                xcorr = xcorr_trace[xcorrpos]
                flip = np.sign(xcorr)
                xcorrelations['{}_ref_xcorr'.format(ref_str)] = xcorr
                xcorrs.append(xcorr)
                xcorrelations['{}_ref_xcorr_time'.format(ref_str)] = xcorrpos * dt

                if self.__debug:
                    if(xcorr > 0.1):
                        fig, (ax, ax2) = plt.subplots(2, 1)
                        ax.set_title('channel {}, xcorr = {:.2f}'.format(channel_id, xcorr))
                        ax.plot(times, trace, label='measurement')
                        tttemp = np.arange(0, len(ref_template_resampled) * dt, dt)
                        ax.plot(tttemp, flip * np.roll(ref_template_resampled * np.abs(trace).max(), xcorrpos), '--', label='template')
                        argmax = np.argmax(trace)
                        ax.set_xlim(argmax * dt - 64 * units.ns, argmax * dt + 64 * units.ns)
                        ax.legend()
                        ax2.plot(hp.get_normalized_xcorr(trace, ref_template_resampled))
                        ax2.set_ylim(-1, 1)
                        plt.tight_layout()
                        plt.show()

            else:
                template_key = []

                for key in ref_templates:

                    ref_template = ref_templates[key][channel.get_id()]
                    ref_template_resampled = self.match_sampling(ref_template, resampling_factor)

                    xcorr_trace = hp.get_normalized_xcorr(trace, ref_template_resampled)
                    xcorrpos = np.argmax(np.abs(xcorr_trace))
                    xcorr = np.abs(xcorr_trace[xcorrpos])

                    xcorrpos_ch.append(xcorrpos)
                    xcorrs_ch.append(xcorr)
                    template_key.append(key)

                if self.__debug:
                    print(event_id)
                    plt.figure()
                    plt.hist(xcorrs_ch, range=(0, 1), bins=50)
                    plt.axvline(np.mean(np.abs(xcorrs_ch)))
                    plt.axvline(np.max(np.abs(xcorrs_ch)))
                    print(np.mean(np.abs(xcorrs_ch)), np.max(np.abs(xcorrs_ch)),
                            channel[chp.maximum_amplitude] / units.mV)

                xcorrelations['{}_ref_xcorr'.format(ref_str)] = np.abs(xcorrs_ch).mean()
                xcorrelations['{}_ref_xcorr_all'.format(ref_str)] = np.abs(xcorrs_ch)
                xcorrelations['{}_ref_xcorr_max'.format(ref_str)] = np.abs(xcorrs_ch[np.argmax(np.abs(xcorrs_ch))])
                xcorrelations['{}_ref_xcorr_time'.format(ref_str)] = np.mean(xcorrpos_ch[np.argmax(np.abs(xcorrs_ch))]) * dt
                xcorrelations['{}_ref_xcorr_template'.format(ref_str)] = template_key[np.argmax(np.abs(xcorrs_ch))]

                logger.debug("average xcorr over all templates {:.2f} +- {:.2f}, \
                    best template is {} at position {:.2f}".format(np.abs(xcorrs_ch).mean(),
                    np.abs(xcorrs_ch).std(),
                    xcorrelations['{}_ref_xcorr_template'.format(ref_str)],
                    xcorrelations['{}_ref_xcorr_time'.format(ref_str)]))

                xcorrs.append(np.nanmean(np.abs(xcorrs_ch)))
                xcorrs_max.append(np.nanmax(np.abs(xcorrs_ch)))

            if self.__debug:
                print("per channel", len(xcorrs_ch))
                plt.hist(xcorrs_ch, range=(0, 1), bins=50)
                plt.show()
                print(xcorrs)
            # Writing information to channel
            if cosmic_ray:
                channel[chp.cr_xcorrelations] = xcorrelations
            else:
                channel[chp.nu_xcorrelations] = xcorrelations

        xcorrs = np.array(xcorrs)
        xcorrs_max = np.array(xcorrs_max)

        if n_templates == 1:

            xcorrelations_station = {}
            xcorrelations_station['number_of_templates'] = n_templates
            xcorrelations_station['{}_max_xcorr'.format(ref_str)] = np.abs(xcorrs).max()
            ref_mask = np.array([channel.get_id() in channels_to_use for channel in station.iter_channels()])
            xcorrelations_station['{0}_max_xcorr_{0}channels'.format(ref_str)] = np.abs(xcorrs[ref_mask]).max()
            xcorrelations_station['{0}_avg_xcorr_{0}channels'.format(ref_str)] = np.abs(xcorrs[ref_mask]).mean()
        else:
            xcorrelations_station = {}
            xcorrelations_station['number_of_templates'] = n_templates
            xcorrelations_station['{}_max_xcorr'.format(ref_str)] = xcorrs_max.max()
            ref_mask = np.array([channel.get_id() in channels_to_use for channel in station.iter_channels()])
            xcorrelations_station['{0}_max_xcorr_{0}channels'.format(ref_str)] = xcorrs_max[ref_mask].max()
            xcorrelations_station['{0}_avg_xcorr_{0}channels'.format(ref_str)] = xcorrs[ref_mask].max()

        # calculate average xcorr in parallel channels
        parallel_channels = det.get_parallel_channels(station_id)
        max_xcorr_parallel = 0
        for pair in parallel_channels:
            mask = np.in1d(pair, channels_to_use)  # we use only the specified channels to calculate the pair averages
            if(np.sum(mask)):
                ref_mask = np.array([channel.get_id() in pair[mask] for channel in station.iter_channels()])
                tmp = np.abs(xcorrs[ref_mask]).mean()
                logger.debug("calculating average xcorr for parallel channels {} = {:.2f}".format(pair[mask], tmp))
                max_xcorr_parallel = max(max_xcorr_parallel, tmp)

        xcorrelations_station['{0}_avg_xcorr_parallel_{0}channels'.format(ref_str)] = max_xcorr_parallel
        logger.debug("best average {0} correlation of parallel {0} channels is\
                     {1:.02f}".format(ref_str,
                     xcorrelations_station['{0}_avg_xcorr_parallel_{0}channels'.format(ref_str)]))

        # Writing information to station
        if cosmic_ray:
            station[stnp.cr_xcorrelations] = xcorrelations_station
        else:
            station[stnp.nu_xcorrelations] = xcorrelations_station

    def end(self):
        pass



import NuRadioReco.modules.io.coreas.readCoREASShower
import NuRadioReco.modules.io.eventWriter
import NuRadioReco.modules.efieldToVoltageConverter
from NuRadioReco.utilities import units
from NuRadioReco.framework.parameters import showerParameters as shP

from NuRadioReco.detector import generic_detector as detector

import numpy as np
import argparse
import datetime

# Parse eventfile as argument
parser = argparse.ArgumentParser(description='NuRadioSim file')
parser.add_argument('inputfilename', type=str, nargs='*',
                    help='path to NuRadioMC simulation result')

parser.add_argument('-o', '--output_filename', type=str, nargs='?',
                    default='Full_CoREAS_shower.nur',
                    help='output file name')

parser.add_argument('--detectordescription', type=str, nargs='?',
                    default='../examples/example_data/dummy_detector.json',
                    help='path to detectordescription')

parser.add_argument('--set_run_number', dest='set_run_number',  action='store_true',
                    help='If set, the run number and event id will be set to an increasing value.')


args = parser.parse_args()

# initialize modules
det = detector.GenericDetector(json_filename=args.detectordescription, default_station=102, default_channel=0)
readCoREASShower = NuRadioReco.modules.io.coreas.readCoREASShower.readCoREASShower()
readCoREASShower.begin(args.inputfilename, det, set_ascending_run_and_event_number=args.set_run_number)
efieldToVoltageConverter = NuRadioReco.modules.efieldToVoltageConverter.efieldToVoltageConverter()
efieldToVoltageConverter.begin()
eventWriter = NuRadioReco.modules.io.eventWriter.eventWriter()
eventWriter.begin(args.output_filename)

for event, gen_det in readCoREASShower.run():
    print('Event {} {}'.format(event.get_run_number(), event.get_id()))
    for station in event.get_stations():
        efieldToVoltageConverter.run(event, station, gen_det)
    eventWriter.run(event, gen_det)

nevents = eventWriter.end()
print("Finished processing, {} events".format(nevents))



import NuRadioReco.modules.channelGenericNoiseAdder
from NuRadioReco.utilities import units
import matplotlib.pyplot as plt
import numpy as np
from scipy import constants
import argparse
channelGenericNoiseAdder = NuRadioReco.modules.channelGenericNoiseAdder.channelGenericNoiseAdder()
from NuRadioReco.utilities.diodeSimulator import diodeSimulator
import NuRadioReco.framework.channel
from NuRadioReco.utilities.fft import time2freq, freq2time
from scipy.signal import butter, freqs

"""
This code calculates the probability per window for the whole phased
array (not just one beam) of triggering on noise, and the rate of
noise triggers.

The trigger is an envelope phased array trigger. According
to the ARA folks, the maximum rate should be around 10 Hz for the whole
phased array.

This code calculates the delays for the beams, introduces noise in all the
phased antennas, calculates the tunnel diode envelope and then phases the array
into the directions given by default_angles. It verifies if there has been a
trigger and repeats.

Increasing the number of samples is equivalent to increasing the number of
tries, since the code calculates the probability of randomly triggering a
single window filled with noise.
"""

parser = argparse.ArgumentParser(description='calculates noise trigger rate for phased array')
parser.add_argument('--ntries', type=int, help='number noise traces to which a trigger is applied for each threshold', default=1000)
parser.add_argument('--single', action='store_true', help='if activated, it calculates the diode noise rate for a single antenna (ARA-like)')
args = parser.parse_args()

main_low_angle = -50. * units.deg
main_high_angle = 50. * units.deg
default_angles = np.arcsin( np.linspace( np.sin(main_low_angle), np.sin(main_high_angle), 30) )

def get_beam_rolls(ant_z, channel_list, phasing_angles=default_angles, time_step=2./3*units.ns, ref_index=1.78):
    """
    Calculates the delays needed for phasing the array.
    """

    beam_rolls = []
    ref_z = (np.max(ant_z)+np.min(ant_z))/2

    n_phased = len(channel_list)

    for angle in phasing_angles:
        subbeam_rolls = {}
        for z, channel_id in zip(ant_z, range(n_phased)):
            delay = (z-ref_z)/0.3 * ref_index * np.sin(angle)
            roll = int(delay/time_step)
            subbeam_rolls[channel_id] = roll
        beam_rolls.append(subbeam_rolls)

        return beam_rolls

min_freq = 132*units.MHz
max_freq = 700*units.MHz
sampling_rate = 3*units.GHz
window_width = int(40*units.ns * sampling_rate)

primary_angles = np.arcsin( np.linspace( np.sin(main_low_angle), np.sin(main_high_angle), 30) )

n_samples = 1000000 # number of samples
time_step = 1./sampling_rate
bandwidth = max_freq-min_freq
amplitude = (300 * 50 * constants.k * bandwidth / units.Hz) ** 0.5

Ntries = args.ntries # number of tries
single = args.single

if not single:
    threshold_factors = [3.8, 3.85, 3.9, 3.95, 4.0]
    ant_z_primary = [-98.5, -99.5, -100.5, -101.5] # primary antennas positions
    primary_channels = [0, 1, 2, 3] # channels used for primary beam
else:
    ant_z_primary = [-98.5]
    primary_channels = [0]
    threshold_factors = [9.85, 9.9, 9.5, 10.]

beam_rolls = get_beam_rolls(ant_z_primary, primary_channels, primary_angles, time_step)

n_beams = len(primary_angles)

passband = (None, 200*units.MHz)
diode = diodeSimulator(passband)
power_mean, power_std = diode.calculate_noise_parameters(sampling_rate,
                                                         min_freq,
                                                         max_freq,
                                                         amplitude=amplitude)

for threshold_factor in threshold_factors:
    prob_per_window = 0

    n_phased = len(primary_channels)
    low_trigger  = power_mean - power_std * np.abs(threshold_factor)
    low_trigger *= n_phased

    for Ntry in range(Ntries):
        noise_array = []

        for iant in range(len(primary_channels)):

            noise_trace = channelGenericNoiseAdder.bandlimited_noise(min_freq, max_freq, n_samples,
                                                sampling_rate, amplitude, type='rayleigh')
            channel = NuRadioReco.framework.channel.Channel(0)
            channel.set_trace(noise_trace, sampling_rate)

            enveloped_noise = diode.tunnel_diode(channel)

            noise_array.append(enveloped_noise)

        for subbeam_rolls in beam_rolls:

            noise = None

            for channel_id in range(len(primary_channels)):

                if(noise is None):
                    noise = np.roll(noise_array[channel_id], subbeam_rolls[channel_id])
                else:
                    noise += np.roll(noise_array[channel_id], subbeam_rolls[channel_id])

            n_windows = int(n_samples/window_width)

            # No overlapping windows, as opposed to the power threshold case
            strides = noise.strides
            windowed_traces = np.lib.stride_tricks.as_strided(noise, \
                              shape=(n_windows, window_width), \
                              strides=(int(window_width) * strides[0], strides[0]))

            threshold_passed = np.any(windowed_traces < low_trigger, axis=1)

            # If a phased direction triggers, the whole phased array triggers.
            # The following formula is justified as long as the probability is small
            # and each direction triggers independently of the rest.
            prob_per_window += np.sum( threshold_passed )/(n_windows*Ntries)

    trigger_frequency = prob_per_window / window_width
    print('Threshold factor: {:.2f}, Fraction of noise triggers: {:.8f}%, Noise trigger rate: {:.2f} Hz'.format(threshold_factor, prob_per_window*100., trigger_frequency/units.Hz))



from NuRadioReco.modules.base.module import register_run
import numpy as np
import copy
import logging
from NuRadioReco.utilities import units
from scipy import signal
logger = logging.getLogger('channelStopFilter')


class channelStopFilter:
    """
    at the beginning and end of the trace (around the 'stop') our electronic produces a glitch
    this modules smoothly filters the beginning and the end of the trace
    """

    def begin(self):
        pass

    @register_run()
    def run(self, evt, station, det, filter_size=0.1, prepend=128 * units.ns, append=128 * units.ns):
        """
        parameters
        ----------
        filter_size: size of tukey window (float)
            specifies the percentage of the trace where the filter is active.
            default is 0.1, i.e. the first 5% and the last 5% of the trace are filtered
        prepend: time interval to prepend
            the time span that is filled with zeros and prepended to the trace
        append: time interval to append
            the time span that is filled with zeros and appended to the trace
        """
        for channel in station.iter_channels():
            trace = channel.get_trace()
            sampling_rate = channel.get_sampling_rate()
            window = signal.tukey(len(trace), filter_size)
            trace *= window
            trace = np.append(np.zeros(np.int(np.round(prepend * sampling_rate))), trace)
            trace = np.append(trace, np.zeros(np.int(np.round(append * sampling_rate))))
            channel.set_trace(trace, sampling_rate)

    def end(self):
        pass



import numpy as np
from scipy import signal
import scipy.signal
import logging

from NuRadioReco.modules.base.module import register_run
from NuRadioReco.utilities import units
from NuRadioReco.detector import filterresponse
import NuRadioReco.framework.sim_station


class channelBandPassFilter:
    """
    Band pass filters the channels using different band-pass filters.
    """
    def __init__(self):
        self.__t = 0
        self.begin()
        self.logger = logging.getLogger('NuRadioReco.channelBandPassFilter')

    def begin(self):
        pass

    @register_run()
    def run(self, evt, station, det, passband=[55 * units.MHz, 1000 * units.MHz],
            filter_type='rectangular', order=2):
        """
        Run the filter

        Parameters
        ---------

        evt, station, det
            Event, Station, Detector
        passband: list
            passband[0]: lower boundary of filter, passband[1]: upper boundary of filter
        filter_type: string
            'rectangular': perfect straight line filter
            'butter': butterworth filter from scipy
            'butterabs': absolute of butterworth filter from scipy
            or any filter that is implemented in NuRadioReco.detector.filterresponse. In this case the
            passband parameter is ignored
            or 'FIR <type> <parameter>' - see below for FIR filter options
        order: int (optional, default 2)
            for a butterworth filter: specifies the order of the filter

        Added Jan-07-2018 by robert.lahmann@fau.de:
        FIR filter:
        see https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.get_window.html
        for window types; some windows need additional parameters; Default window is hamming
        (does not need parameters). Examples:
        filter_type='FIR' : use hamming window
        filter_type='FIR hamming': same as above
        filter_type='FIR kaiser 10' : Use Kaiser window with beta parameter 10
        filter_type='FIR kaiser' : Use Kaiser window with default beta parameter 6

        In principal, window names are just passed on to signal.firwin(), but if parameters
        are required, then these cases must be explicitly implemented in the code below.

        The four main filter types can be implemented:
        LP: passband[0]=None, passband[1]  = f_cut
        HP: passband[0]=f_cut, passband[1] = None
        BP: passband[0]=f_cut_low, passband[1] = f_cut_high
        BS: passband[0]=f_cut_high, passband[1] = f_cut_low (i.e. passband[0] > passband[1])

        """

        for channel in station.iter_channels():
            self._apply_filter(channel, passband, filter_type, order, False)

    def get_filter(self, frequencies, station_id, channel_id, det, passband, filter_type, order=2):
        """
        helper function to return the filter that the module applies.

        The unused parameters station_id, channel_id, det are needed to have the same signature as the
        `get_filter` functions of other modules, e.g. the hardwareResponseIncorporator.

        Parameters
        -----------
        frequencies: array of floats
            the frequency array for which the filter should be returned
        station_id: int
            the station id
        channel_id: int
            the channel id
        det: detector instance
            the detector
        the remaining arguments are the arguments of the run method!

        Returns: array of complex floats
            the complex filter amplitudes
        """
        if(filter_type == 'rectangular'):
            f = np.ones_like(frequencies)
            f[np.where(frequencies < passband[0])] = 0.
            f[np.where(frequencies > passband[1])] = 0.
            return f
        elif(filter_type == 'butter'):
            f = np.zeros_like(frequencies, dtype=np.complex)
            mask = frequencies > 0
            b, a = scipy.signal.butter(order, passband, 'bandpass', analog=True)
            w, h = scipy.signal.freqs(b, a, frequencies[mask])
            f[mask] = h
            return f
        elif(filter_type == 'butterabs'):
            f = np.zeros_like(frequencies, dtype=np.complex)
            mask = frequencies > 0
            b, a = scipy.signal.butter(order, passband, 'bandpass', analog=True)
            w, h = scipy.signal.freqs(b, a, frequencies[mask])
            f[mask] = h
            return np.abs(f)
        elif(filter_type.find('FIR') >= 0):
            raise NotImplementedError("FIR filter not yet implemented")
        else:
            return filterresponse.get_filter_response(frequencies, filter_type)

    def _apply_filter(self, channel, passband, filter_type, order, is_efield=False):
#         print(f"apply_filter self {self}")
#         print(channel)

        frequencies = channel.get_frequencies()
        trace_fft = channel.get_frequency_spectrum()
        sample_rate = channel.get_sampling_rate()

        # for FIR filters, it is easier to set the trace rather than the FFT to apply the
        # filter response. Eventually, I (Robert Lahmann) might figure out a way to set the
        # FFT (as is done for the IIR filters corresponding to analog filters) but for now
        # I am setting a variable to decide whether the FFT or time trace shall be used to
        # set the trace.

        isFIR = False

        if(filter_type == 'rectangular'):
            trace_fft *= self.get_filter(frequencies, 0, 0, None, passband, filter_type)
        elif(filter_type == 'butter'):
            trace_fft *= self.get_filter(frequencies, 0, 0, None, passband, filter_type, order)
        elif(filter_type == 'butterabs'):
            trace_fft *= self.get_filter(frequencies, 0, 0, None, passband, filter_type, order)
        elif(filter_type.find('FIR') >= 0):
            # print('This is a FIR filter')
            firarray = filter_type.split()
            if (len(firarray) == 1):
                wtype = 'hamming'
                # print('hamming window')
            else:
                wtype = firarray[1]
                if wtype.find('kaiser') >= 0:
                    if len(firarray) > 2:
                        beta = float(firarray[2])
                    else:
                        beta = 6.0
                    wtype = ('kaiser', beta)
            # print('window type: ', wtype)
            Nfir = order + 1
            if (passband[0] == None):
                # this is a low pass filter
                pass_zero = True
                fcut = passband[1]
            elif (passband[1] == None or passband[1] / sample_rate >= 0.5):
                # this is a high pass filter
                pass_zero = False
                fcut = passband[0]
            elif (passband[1] > passband[0]):
                # this is a bandpass filter
                pass_zero = False
                fcut = passband
            elif(passband[0] > passband[1]):
                # this is a bandstop filter
                pass_zero = True
                fcut = [passband[1], passband[0]]
                # print('bandstop with fcut = ',fcut)
            else:
                # something went wrong!!
                print("Error, could not define filter type")
            # print('fcut = ',fcut)
            taps = signal.firwin(Nfir, fcut, window=wtype, scale=False, pass_zero=pass_zero, fs=sample_rate)
            wfilt, hfilt = signal.freqz(taps, worN=len(frequencies))

            if ((Nfir // 2) * 2 == Nfir):
                print("odd filter order, rolling is off by T_s/2")

            ndelay = int(0.5 * (Nfir - 1))
            trace_fir = signal.lfilter(taps, 1.0, channel.get_trace())
            # print('len(trace_fir)',len(trace_fir))
            trace_fir = np.roll(trace_fir, -ndelay)
#             channel.set_trace(trace_fir, sample_rate)
#             #channel.set_trace(trace_fir, sample_rate)
#             #trace_fft = hfilt
#             trace_fft = channel.get_frequency_spectrum()
            isFIR = True
        else:
            trace_fft *= self.get_filter(frequencies, 0, 0, None, passband, filter_type)
        if isFIR:
            channel.set_trace(trace_fir, sample_rate)
            # print('set trace for fir')
        else:
            channel.set_frequency_spectrum(trace_fft, sample_rate)

    def end(self):
        pass



from NuRadioReco.modules.base.module import register_run
from NuRadioReco.utilities import units
from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.trigger import HighLowTrigger
import numpy as np
import time
import logging
logger = logging.getLogger('HighLowTriggerSimulator')


def get_high_low_triggers(trace, high_threshold, low_threshold,
                          time_coincidence=5 * units.ns, dt=1 * units.ns):
    """
    calculats a high low trigger in a time coincidence window

    Parameters
    ----------
    trace: array of floats
        the signal trace
    high_threshold: float
        the high threshold
    low_threshold: float
        the low threshold
    time_coincidence: float
        the time coincidence window between a high + low
    dt: float
        the width of a time bin (inverse of sampling rate)

    Returns
    -------
    triggered bins: array of bools
        the bins where the trigger condition is satisfied
    """
    n_bins_coincidence = np.int(np.round(time_coincidence / dt)) + 1
    c = np.ones(n_bins_coincidence, dtype=np.bool)
    logger.debug("length of trace {} bins, coincidence window {} bins".format(len(trace), len(c)))

    c2 = np.array([1, -1])
    m1 = np.convolve(trace > high_threshold, c, mode='full')[:-(n_bins_coincidence - 1)]
    m2 = np.convolve(trace < low_threshold, c, mode='full')[:-(n_bins_coincidence - 1)]
    return np.convolve(m1 & m2, c2, mode='same') > 0


def get_majority_logic(tts, number_of_coincidences=2, time_coincidence=32 * units.ns, dt=1 * units.ns):
    """
    calculates a majority logic trigger

    Parameters
    ----------
    tts: array/list of array of bools
        an array of bools that indicate a single channel trigger per channel
    number_of_coincidences: int (default: 2)
        the number of coincidences between channels
    time_coincidence: float
        the time coincidence window between channels
    dt: float
        the width of a time bin (inverse of sampling rate)

    Returns:
    --------
    triggerd: bool
        returns True if majority logic is fulfilled
    triggerd_bins: array of ints
        the bins that fulfilled the trigger
    triggered_times: array of floats
        the trigger times
    """
    n = len(tts[0])
    n_bins_coincidence = np.int(np.round(time_coincidence / dt)) + 1
    if(n_bins_coincidence > n):  # reduce coincidence window to maximum trace length
        n_bins_coincidence = n
        logger.debug("specified coincidence window longer than tracelenght, reducing coincidence window to trace length")
    c = np.ones(n_bins_coincidence, dtype=np.bool)

    for i in range(len(tts)):
        logger.debug("get_majority_logic() length of trace {} bins, coincidence window {} bins".format(len(tts[i]), len(c)))
        tts[i] = np.convolve(tts[i], c, mode='full')[:-(n_bins_coincidence - 1)]
    tts = np.sum(tts, axis=0)
    ttt = tts >= number_of_coincidences
    triggered_bins = np.atleast_1d(np.squeeze(np.argwhere(tts >= number_of_coincidences)))
    return np.any(ttt), triggered_bins, triggered_bins * dt


class triggerSimulator:
    """
    Calculates the trigger of an event.
    Uses the ARIANNA trigger logic, that a single antenna needs to cross a high and a low threshold value,
    and then coincidences between antennas can be required.
    """

    def __init__(self):
        self.__t = 0
        self.begin()

    def begin(self, log_level=None):
        if(log_level is not None):
            logger.setLevel(log_level)
        return

    @register_run()
    def run(self, evt, station, det,
            threshold_high=60 * units.mV,
            threshold_low=-60 * units.mV,
            high_low_window=5 * units.ns,
            coinc_window=200 * units.ns,
            number_concidences=2,
            triggered_channels=None,
            trigger_name="default_high_low",
            set_not_triggered=False):
        """
        simulate ARIANNA trigger logic

        Parameters
        ----------
        threshold_high: float
            the threshold voltage that needs to be crossed on a single channel on the high side
        threshold_low: float
            the threshold voltage that needs to be crossed on a single channel on the low side
        high_low_window: float
           time window in which a high+low crossing needs to occur to trigger a channel
        coinc_window: float
            time window in which number_concidences channels need to trigger
        number_concidences: int
            number of channels that are requried in coincidence to trigger a station
        triggered_channels: array of ints or None
            channels ids that are triggered on, if None trigger will run on all channels
        cut_trace: bool
            if true, trace is cut to the correct length (50ns before the trigger,
            max trace length is set according to detector description)
        trigger_name: string
            a unique name of this particular trigger
        set_not_triggered: bool (default: False)
            if True not trigger simulation will be performed and this trigger will be set to not_triggered

        """
        t = time.time()
        if threshold_low >= threshold_high:
            logger.error("Impossible trigger configuration, high {0} low {1}.".format(threshold_high, threshold_low))
            raise NotImplementedError

        sampling_rate = station.get_channel(0).get_sampling_rate()
        channels_that_passed_trigger = []
        if not set_not_triggered:
            max_signal = 0

            triggerd_bins_channels = []
            dt = 1. / sampling_rate
            if triggered_channels is None:
                for channel in station.iter_channels():
                    channel_trace_start_time = channel.get_trace_start_time()
                    break
            else:
                channel_trace_start_time = station.get_channel(triggered_channels[0]).get_trace_start_time()
            for channel in station.iter_channels():
                channel_id = channel.get_id()
                if triggered_channels is not None and channel_id not in triggered_channels:
                    continue
                if channel.get_trace_start_time() != channel_trace_start_time:
                    logger.warning('Channel has a trace_start_time that differs from the other channels. The trigger simulator may not work properly')
                trace = channel.get_trace()
                triggerd_bins = get_high_low_triggers(trace, threshold_high, threshold_low,
                                                      high_low_window, dt)
                if True in triggerd_bins:
                    channels_that_passed_trigger.append(channel.get_id())
                triggerd_bins_channels.append(triggerd_bins)
                logger.debug("channel {}, len(triggerd_bins) = {}".format(channel_id, len(triggerd_bins)))

            has_triggered, triggered_bins, triggered_times = get_majority_logic(
                triggerd_bins_channels, number_concidences, coinc_window, dt)
            # set maximum signal aplitude
            max_signal = 0
            if(has_triggered):
                for channel in station.iter_channels():
                    max_signal = max(max_signal, np.abs(channel.get_trace()[triggered_bins]).max())
                station.set_parameter(stnp.channels_max_amplitude, max_signal)
        else:
            logger.info("set_not_triggered flag True, setting triggered to False.")
            has_triggered = False

        trigger = HighLowTrigger(trigger_name, threshold_high, threshold_low, high_low_window,
                                 coinc_window, channels=triggered_channels, number_of_coincidences=number_concidences)
        trigger.set_triggered_channels(channels_that_passed_trigger)
        if not has_triggered:
            trigger.set_triggered(False)
            logger.info("Station has NOT passed trigger")
            trigger.set_trigger_time(0)
        else:
            trigger.set_triggered(True)
            trigger.set_trigger_time(triggered_times.min()+channel_trace_start_time)
            trigger.set_trigger_times(triggered_times+channel_trace_start_time)
            logger.info("Station has passed trigger, trigger time is {:.1f} ns".format(
                trigger.get_trigger_time() / units.ns))

        station.set_trigger(trigger)

        self.__t += time.time() - t

    def end(self):
        from datetime import timedelta
        logger.setLevel(logging.INFO)
        dt = timedelta(seconds=self.__t)
        logger.info("total time used by this module is {}".format(dt))
        return dt



#!/usr/bin/env python
import NuRadioReco.detector.detector
import NuRadioReco.modules.io.eventReader
import NuRadioReco.modules.io.eventWriter
import NuRadioReco.modules.ARIANNA.hardwareResponseIncorporator
import NuRadioReco.modules.trigger.highLowThreshold
import NuRadioReco.modules.trigger.multiHighLowThreshold
import NuRadioReco.modules.trigger.simpleThreshold
import NuRadioReco.modules.phasedarray.triggerSimulator
import NuRadioReco.modules.efieldToVoltageConverter
from NuRadioReco.utilities import units
import datetime

det = NuRadioReco.detector.detector.Detector(json_filename='NuRadioReco/test/trigger_tests/trigger_test_detector.json')
det.update(datetime.datetime(2018, 10, 1))
event_reader = NuRadioReco.modules.io.eventReader.eventReader()
event_reader.begin('NuRadioReco/test/trigger_tests/trigger_test_input.nur')
event_writer = NuRadioReco.modules.io.eventWriter.eventWriter()
event_writer.begin('NuRadioReco/test/trigger_tests/trigger_test_output.nur')

high_low_trigger = NuRadioReco.modules.trigger.highLowThreshold.triggerSimulator()
multi_high_low_trigger = NuRadioReco.modules.trigger.multiHighLowThreshold.triggerSimulator()
simple_threshold_trigger = NuRadioReco.modules.trigger.simpleThreshold.triggerSimulator()
phased_array_trigger = NuRadioReco.modules.phasedarray.triggerSimulator.triggerSimulator()
efield_to_voltage_converter = NuRadioReco.modules.efieldToVoltageConverter.efieldToVoltageConverter()
efield_to_voltage_converter.begin()
hardware_response_incorporator = NuRadioReco.modules.ARIANNA.hardwareResponseIncorporator.hardwareResponseIncorporator()

for event in event_reader.run():
    station = event.get_station(1)
    efield_to_voltage_converter.run(event, station, det)
    hardware_response_incorporator.run(event, station, det, True)
    high_low_trigger.run(event, station, det, threshold_high=40 * units.mV, threshold_low=-40 * units.mV)
    multi_high_low_trigger.run(event, station, det, trigger_name="default_multi_high_low", threshold_high=40 * units.mV, threshold_low=-40 * units.mV, n_high_lows=2)
    simple_threshold_trigger.run(event, station, det)
    phased_array_trigger.run(event, station, det, threshold=40 * units.mV,
                             secondary_channels=[0,1,3,4,6,7])

    event_writer.run(event)



import numpy as np
from NuRadioReco.utilities import units
from radiotools import helper as hp
from scipy import interpolate as intp
from scipy import integrate as int
from matplotlib import pyplot as plt
import os
import glob


def get_amp_response(frequencies, amp_name):
    """
    Get measured amplifier responses.
    Very simple script for currently NTU amps.

    """
    # parse input file and convert to default units
    directory = os.path.dirname(os.path.abspath(__file__))
    ff2, phase, t = np.loadtxt(os.path.join(
        directory, 'amps/{}/LNA_PHASE.CSV'.format(amp_name)), unpack=True, skiprows=3, delimiter=',')
    ff, logmag, t = np.loadtxt(os.path.join(
        directory, 'amps/{}/LNA_LOGMAG.CSV'.format(amp_name)), unpack=True, skiprows=3, delimiter=',')
    ff *= units.Hz
    ff2 *= units.Hz
    phase *= units.deg
    linmag = 10**(logmag/20.)

    get_phase = intp.interp1d(ff2, np.unwrap(phase))
    get_linmag = intp.interp1d(ff, linmag)

    response = np.zeros_like(frequencies, dtype=np.complex)
    mask = (frequencies > max(ff.min(), ff2.min())) & (frequencies < min(ff.max(), ff2.max()))
    response[mask] = get_linmag(frequencies[mask]) * np.exp(1j * get_phase(frequencies[mask]))
    return response

if __name__ == "__main__":
    ff = np.linspace(10 * units.MHz, 500 * units.MHz, 100)
    df = ff[1] - ff[0]
    fig, (ax, ax2) = plt.subplots(1, 2)
    response = get_amp_response(ff, "NTU01")
    ax.plot(ff / units.MHz, np.abs(response))
    group_delay = -1./2/np.pi * np.diff(np.unwrap(np.angle(response))) / df
    #     ax2.plot(ff / units.MHz, np.unwrap(np.angle(response1)) / units.deg)
    ax2.plot((ff[1:] + ff[:-1]) * 0.5 / units.MHz, group_delay/units.ns)
    ax.legend()
    ax2.legend()
    fig.tight_layout()
    plt.show()


from NuRadioReco.modules.base.module import register_run
import ROOT
import random
from NuRadioReco.utilities import units
import numpy as np
import logging
logger = logging.getLogger('noiseImporter')


class noiseImporter:
    """
    Imports recorded noise from ARIANNA station. The recorded noise needs to match the station geometry and sampling
    as chosen with channelResampler and channelLengthAdjuster
    
    For different stations, new noise files need to be used. 
    Collect forced triggers from any type of station to use for analysis.
    A seizable fraction of data is recommended for accuracy. 
    
    The noise will be random. This module therefore might produce non-reproducible results on a single event basis,
    if run several times.  
    """

    def begin(self, noise_files):

        self.nt = ROOT.TChain("CalibTree")
        if len(noise_files) > 1:
            logger.warning("Only using the first noise file, more is not implemented yet")
        self.nt.Add(noise_files[0])

        self.data = ROOT.TSnCalWvData()
        self.nt.SetBranchAddress("AmpOutData.", self.data)

        self.nevts = self.nt.GetEntries()

    @register_run()
    def run(self, evt, station, det):
        # loop over stations in simulation
        for channel in station.iter_channels():
            channel_id = channel.get_id()

            # pick a noise waveform
            noise_event = random.randint(0, self.nevts)
            self.nt.GetEntry(noise_event)
#                 nchans = self.data.GetNumChans()
            noise_samples = self.data.GetNumSamplesOn(0)

            trace = channel.get_trace()
            if (noise_samples != trace.shape[0]):
                logger.warning("Mismatch: Noise has {0} and simulation {1} samples\n Not adding noise!".format(noise_samples, trace.shape[0]))
            else:
                noise_trace = np.array(self.data.GetDataOnCh(channel_id)) * units.mV
                noise_trace += trace

                channel.set_trace(noise_trace, channel.get_sampling_rate())

    def end(self):
        pass



from NuRadioReco.modules.base.module import register_run
import numpy as np
import os
import copy
from NuRadioReco.utilities import geometryUtilities as geo_utl
from NuRadioReco.utilities import units, fft
from NuRadioReco.utilities import ice
from NuRadioReco.detector import antennapattern
from NuRadioReco.utilities import trace_utilities
import NuRadioReco.framework.base_trace
import NuRadioReco.framework.electric_field
import matplotlib.pyplot as plt
from scipy import signal

import logging
logger = logging.getLogger('voltageToEfieldConverter')

from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.parameters import channelParameters as chp
from NuRadioReco.framework.parameters import electricFieldParameters as efp


def get_array_of_channels(station, use_channels, det, zenith, azimuth,
                          antenna_pattern_provider, time_domain=False):
    time_shifts = np.zeros(len(use_channels))
    t_cables = np.zeros(len(use_channels))
    t_geos = np.zeros(len(use_channels))

    station_id = station.get_id()
    site = det.get_site(station_id)
    n_ice = ice.get_refractive_index(-0.01, site)
    for iCh, channel in enumerate(station.iter_channels(use_channels)):
        channel_id = channel.get_id()

        antenna_position = det.get_relative_position(station_id, channel_id)
        # determine refractive index of signal propagation speed between antennas
        refractive_index = ice.get_refractive_index(1, site)  # if signal comes from above, in-air propagation speed
        if(zenith > 0.5 * np.pi):
            refractive_index = ice.get_refractive_index(antenna_position[2], site)  # if signal comes from below, use refractivity at antenna position
        time_shift = -geo_utl.get_time_delay_from_direction(zenith, azimuth, antenna_position, n=refractive_index)
        t_geos[iCh] = time_shift
        t_cables[iCh] = channel.get_trace_start_time()
        logger.debug("time shift channel {}: {:.2f}ns (signal prop), {:.2f}ns (trace start time)".format(channel.get_id(), time_shift, channel.get_trace_start_time()))
        time_shift += channel.get_trace_start_time()
        time_shifts[iCh] = time_shift

    delta_t = time_shifts.max() - time_shifts.min()
    tmin = time_shifts.min()
    tmax = time_shifts.max()
    logger.debug("adding relative station time = {:.0f}ns".format((t_cables.min() + t_geos.max()) / units.ns))
    logger.debug("delta t is {:.2f}".format(delta_t / units.ns))
    trace_length = station.get_channel(0).get_times()[-1] - station.get_channel(0).get_times()[0]
    debug_cut = 0
    if(debug_cut):
        fig, ax = plt.subplots(len(use_channels), 1)

    traces = []
    n_samples = None
    for iCh, channel in enumerate(station.iter_channels(use_channels)):
        tstart = delta_t - (time_shifts[iCh] - tmin)
        tstop = tmax - time_shifts[iCh] - delta_t + trace_length
        iStart = int(round(tstart * channel.get_sampling_rate()))
        iStop = int(round(tstop * channel.get_sampling_rate()))
        if(n_samples is None):
            n_samples = iStop - iStart
            if(n_samples % 2):
                n_samples -= 1

        trace = copy.copy(channel.get_trace())  # copy to not modify data structure
        trace = trace[iStart:(iStart + n_samples)]
        if(debug_cut):
            ax[iCh].plot(trace)
        base_trace = NuRadioReco.framework.base_trace.BaseTrace()  # create base trace class to do the fft with correct normalization etc.
        base_trace.set_trace(trace, channel.get_sampling_rate())
        traces.append(base_trace)
    times = traces[0].get_times()  # assumes that all channels have the same sampling rate
    if(time_domain):  # save time domain traces first to avoid extra fft
        V_timedomain = np.zeros((len(use_channels), len(times)))
        for iCh, trace in enumerate(traces):
            V_timedomain[iCh] = trace.get_trace()
    frequencies = traces[0].get_frequencies()  # assumes that all channels have the same sampling rate
    V = np.zeros((len(use_channels), len(frequencies)), dtype=np.complex)
    for iCh, trace in enumerate(traces):
        V[iCh] = trace.get_frequency_spectrum()

    efield_antenna_factor = trace_utilities.get_efield_antenna_factor(station, frequencies, use_channels, det, zenith, azimuth, antenna_pattern_provider)

    if(debug_cut):
        plt.show()

    if(time_domain):
        return efield_antenna_factor, V, V_timedomain

    return efield_antenna_factor, V


def stacked_lstsq(L, b, rcond=1e-10):
    """
    Solve L x = b, via SVD least squares cutting of small singular values
    L is an array of shape (..., M, N) and b of shape (..., M).
    Returns x of shape (..., N)
    """
    u, s, v = np.linalg.svd(L, full_matrices=False)
    s_max = s.max(axis=-1, keepdims=True)
    s_min = rcond * s_max
    inv_s = np.zeros_like(s)
    inv_s[s >= s_min] = 1 / s[s >= s_min]
    x = np.einsum('...ji,...j->...i', v,
                  inv_s * np.einsum('...ji,...j->...i', u, b.conj()))
    return np.conj(x, x)


class voltageToEfieldConverter:

    def __init__(self):
        self.begin()

    def begin(self):
        self.antenna_provider = antennapattern.AntennaPatternProvider()
        pass

    @register_run()
    def run(self, evt, station, det, debug=False, debug_plotpath=None, use_channels=[0, 1, 2, 3], use_MC_direction=False):
        """
        run method. This function is executed for each event

        Parameters
        ---------
        evt
        station
        det
        debug: bool
            if True debug plotting is enables
        debug_plotpath: string or None
            if not None plots will be saved to a file rather then shown. Plots will
            be save into the `debug_plotpath` directory
        use_channels: array of ints
            the channel ids to use for the electric field reconstruction
        """
        event_time = station.get_station_time()
        station_id = station.get_id()

        if use_MC_direction:
            zenith = station.get_sim_station()[stnp.zenith]
            azimuth = station.get_sim_station()[stnp.azimuth]
            sim_present = True
        else:
            logger.info("Using reconstructed (or starting) angles as no signal arrival angles are present")
            zenith = station[stnp.zenith]
            azimuth = station[stnp.azimuth]
            sim_present = False

        efield_antenna_factor, V = get_array_of_channels(station, use_channels, det, zenith, azimuth, self.antenna_provider)
        n_frequencies = len(V[0])
        denom = (efield_antenna_factor[0][0] * efield_antenna_factor[1][1] - efield_antenna_factor[0][1] * efield_antenna_factor[1][0])
        mask = np.abs(denom) != 0
        # solving for electric field using just two orthorgonal antennas
        E1 = np.zeros_like(V[0])
        E2 = np.zeros_like(V[0])
        E1[mask] = (V[0] * efield_antenna_factor[1][1] - V[1] * efield_antenna_factor[0][1])[mask] / denom[mask]
        E2[mask] = (V[1] - efield_antenna_factor[1][0] * E1)[mask] / efield_antenna_factor[1][1][mask]
        denom = (efield_antenna_factor[0][0] * efield_antenna_factor[-1][1] - efield_antenna_factor[0][1] * efield_antenna_factor[-1][0])
        mask = np.abs(denom) != 0
        E1[mask] = (V[0] * efield_antenna_factor[-1][1] - V[-1] * efield_antenna_factor[0][1])[mask] / denom[mask]
        E2[mask] = (V[-1] - efield_antenna_factor[-1][0] * E1)[mask] / efield_antenna_factor[-1][1][mask]
        # solve it in a vectorized way
        efield3_f = np.zeros((2, n_frequencies), dtype=np.complex)
        efield3_f[:, mask] = np.moveaxis(stacked_lstsq(np.moveaxis(efield_antenna_factor[:, :, mask], 2, 0), np.moveaxis(V[:, mask], 1, 0)), 0, 1)
        # add eR direction
        efield3_f = np.array([np.zeros_like(efield3_f[0], dtype=np.complex),
                             efield3_f[0],
                             efield3_f[1]])

        electric_field = NuRadioReco.framework.electric_field.ElectricField(use_channels, [0, 0, 0])
        electric_field.set_frequency_spectrum(efield3_f, station.get_channel(0).get_sampling_rate())
        electric_field.set_parameter(efp.zenith, zenith)
        electric_field.set_parameter(efp.azimuth, azimuth)
        # figure out the timing of the E-field
        t_shifts = np.zeros(V.shape[0])
        site = det.get_site(station_id)
        if(zenith > 0.5 * np.pi):
            logger.warning("Module has not been optimized for neutrino reconstruction yet. Results may be nonsense.")
            refractive_index = ice.get_refractive_index(-1, site)  # if signal comes from below, use refractivity at antenna position
        else:
            refractive_index = ice.get_refractive_index(1, site)  # if signal comes from above, in-air propagation speed
        for i_ch, channel_id in enumerate(use_channels):
            antenna_position = det.get_relative_position(station.get_id(), channel_id)
            t_shifts[i_ch] = station.get_channel(channel_id).get_trace_start_time() - geo_utl.get_time_delay_from_direction(zenith, azimuth, antenna_position, n=refractive_index)

        electric_field.set_trace_start_time(t_shifts.max())
        station.add_electric_field(electric_field)

    def end(self):
        pass



from __future__ import absolute_import, division, print_function, unicode_literals
from NuRadioReco.modules.base.module import register_run
import os
import time
import numpy as np
from numpy.polynomial import polynomial as poly
from scipy import signal
from scipy.signal import correlate
from scipy import optimize as opt
import matplotlib.pyplot as plt

from radiotools import plthelpers as php
from radiotools import helper as hp
from radiotools import coordinatesystems

from NuRadioReco.detector import detector
from NuRadioReco.detector import antennapattern
from NuRadioReco.utilities import geometryUtilities as geo_utl
from NuRadioReco.utilities import units, fft, trace_utilities
from NuRadioReco.utilities import analytic_pulse as pulse
from NuRadioReco.modules.voltageToEfieldConverter import get_array_of_channels

from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.parameters import channelParameters as chp
from NuRadioReco.framework.parameters import electricFieldParameters as efp
import NuRadioReco.framework.electric_field

import logging
logger = logging.getLogger('voltageToAnalyticEfieldConverter')


def covariance(function, vmin, up, fast=False):
    """
    Numerically compute the covariance matrix from a chi^2 or -logLikelihood function.

    Parameters
    ----------
    function: function-like
      The function may accept only a vector argument and has to return a scalar.
    vmin: array of floats
      Position of the minimum.
    up: float
      Threshold value to pass when climbing uphill.
      up = 1   for a chi^2 function
      up = 0.5 for a -logLikelihood function
    fast: boolean
      If true invert hesse matrix at the minimum, use this if computing function is expensive.

    Examples
    --------
    >>> cov = ((2.0,0.2),(0.2,2.0))
    >>> invcov = np.linalg.inv(cov)
    >>> xs = np.array((1.0,-1.0))
    >>> def ChiSquare(pars, grad = None): return np.dot(xs-pars,np.dot(invcov,xs-pars))
    >>> def NegLogLike(pars, grad = None): return 0.5*ChiSquare(pars)
    >>> covariance(ChiSquare, xs, 1.0)
    array([[ 2. ,  0.2],
           [ 0.2,  2. ]])
    >>> covariance(ChiSquare, xs, 1.0, fast=True)
    array([[ 2. ,  0.2],
           [ 0.2,  2. ]])
    >>> covariance(NegLogLike, xs, 0.5)
    array([[ 2. ,  0.2],
           [ 0.2,  2. ]])
    >>> covariance(NegLogLike, xs, 0.5, fast=True)
    array([[ 2. ,  0.2],
           [ 0.2,  2. ]])

    Notes
    -----
    The algorithm is slow (it needs many function evaluations), but robust.
    The covariance matrix is derived by explicitly following the chi^2
    or -logLikelihood function uphill until it crosses the 1-sigma contour.

    The fast alternative is to invert the hessian matrix at the minimum.

    Author
    ------
    Hans Dembinski <hans.dembinski@kit.edu>
    """

    from scipy.optimize import brentq

    class Func:

        def __init__(self, function, vmin, up):
            self.dir = np.zeros_like(vmin)
            self.up = up
            self.vmin = vmin
            self.fmin = function(vmin)
            self.func = function

        def __call__(self, x):
            return self.func(self.vmin + x * self.dir) - self.fmin - self.up

        def SetDirection(self, i, j):
            self.dir *= 0
            self.dir[abs(i)] = 1 if i >= 0 else -1
            self.dir[abs(j)] = 1 if j >= 0 else -1

        def GetBoundary(self, sign):
            eps = np.sqrt(np.finfo(np.double).eps)
            h = eps
            x0 = abs(np.dot(self.vmin, self.dir))

            def IsNonsense(x):
                return np.isnan(x) or np.isinf(x)

            def x(h):
                return sign * (h * x0 if x0 != 0 else h)

            while True:
                # (1) do smallest possible step first,
                #     then grow exponentially until zero+ is crossed,

                if IsNonsense(x(h)):
                    raise StandardError("profile does not cross fmin + up")

                t = self(x(h))

                if IsNonsense(t):
                    # (2) if stepped into nonsense region (infinite, nan, ...),
                    #     do bisection search towards last valid step
                    a = h / 8.0
                    b = h
                    while True:
                        if 2 * (b - a) < eps * (b + a):
                            raise StandardError("profile does not cross fmin + up")
                    h = (a + b) / 2.0
                    t = self(x(h))

                    if IsNonsense(t):
                        b = h
                        continue

                    if t < 0:
                        a = h
                        continue

                    return x(h)

                if t > 0:
                    return x(h)

                h *= 16

    n = len(vmin)

    if fast:
        from pyik.numpyext import hessian
        releps = 1e-3
        dvmin = vmin * releps
        dvmin[dvmin == 0] = releps
        a = hessian(function, vmin, dvmin) / up
    else:
        # Ansatz: (f(r) - fmin)/up = 1/2 r^T C r == 1
        # Diagonal elements:
        # 1 != 1/2 sum_{ij} delta_ik x delta_jk x C_ij
        #    = x^2/2 C_kk
        # => C_kk = 2/x^2
        # Off-diagonal elements:
        # 1 != 1/2 x (delta_ik + delta_il) C_ij x (delta_jk + delta_jl)
        #    = x^2/2 (C_kk + C_kl + C_lk + C_ll) = x^2/2 (2 C_kl + C_kk + C_ll)
        # => C_kl = 0.5 * (2/x^2 - C_kk - C_ll)

        func = Func(function, vmin, up)
        d = np.empty((n, n))
        for i in xrange(n):
            func.SetDirection(i, i)

            xu = func.GetBoundary(+1)
            t = func(-xu)
            xd = -xu if not np.isinf(t) and t > 0.0 else func.GetBoundary(-1)

            x1 = +brentq(func, 0, xu)
            x2 = -brentq(func, xd, 0)
            x = 0.5 * (x1 + x2)

            if x < 0:
                raise StandardError("x may not be negative")

            d[i, i] = x

        for i in xrange(n - 1):
            for j in xrange(i + 1, n):
                func.SetDirection(i, j)

                xu = func.GetBoundary(+1)
                t = func(-xu)
                xd = -xu if not np.isinf(t) and t > 0.0 else func.GetBoundary(-1)

                x1 = +brentq(func, 0, xu)
                x2 = -brentq(func, xd, 0)
                x = 0.5 * (x1 + x2)

                if x < 0:
                    raise StandardError("x may not be negative")

                # check whether x is in possible range
                a = d[i, i]
                b = d[j, j]
                xmax = np.inf if a <= b else 1.0 / (1.0 / b - 1.0 / a)
                xmin = 1.0 / (1.0 / b + 1.0 / a)

                if x <= xmin:
                    logger.warning("covariance(...):", xmin, "<", x, "<", xmax, "violated")
                    x = xmin * 1.01
                if x >= xmax:
                    logger.warning("covariance(...):", xmin, "<", x, "<", xmax, "violated")
                    x = xmax * 0.99

                d[i, j] = d[j, i] = x

        a = 2.0 / d ** 2

        for i in xrange(n - 1):
            for j in xrange(i + 1, n):
                a[i, j] = a[j, i] = 0.5 * (a[i, j] - a[i, i] - a[j, j])

    # Beware: in case of a chi^2 we calculated
    # t^2 = (d^2 chi^2 / d par^2)^{-1},
    # while s^2 = (1/2 d^2 chi^2 / d par^2)^{-1} is correct,
    # thus s^2 = 2 t^2
    cov = 2.0 * np.linalg.inv(a)

    # first aid, if 1-sigma contour does not look like hyper-ellipsoid
    for i in xrange(n):
        if cov[i, i] < 0:
            logger.warning("covariance(...): error, cov[%i,%i] < 0, returning zero" % (i, i))
            for j in xrange(n):
                cov[i, j] = 0

    return cov


def stacked_lstsq(L, b, rcond=1e-10):
    """
    Solve L x = b, via SVD least squares cutting of small singular values
    L is an array of shape (..., M, N) and b of shape (..., M).
    Returns x of shape (..., N)
    """
    u, s, v = np.linalg.svd(L, full_matrices=False)
    s_max = s.max(axis=-1, keepdims=True)
    s_min = rcond * s_max
    inv_s = np.zeros_like(s)
    inv_s[s >= s_min] = 1 / s[s >= s_min]
    x = np.einsum('...ji,...j->...i', v,
                  inv_s * np.einsum('...ji,...j->...i', u, b.conj()))
    return np.conj(x, x)


class voltageToAnalyticEfieldConverter:
    """
    reconstucts the electric-field by foward folding an analytic pulse function through the antenna

    This module works only for cosmic rays so far. The cosmic-ray radio pulse
    can be described in Fourier space with a simple exponential function for
    the magnitude as a function of frequency.
    The phase is set to zero. A slope in the phase spectrum corresponds to a
    translation in time. In each iteration step, the time shift is set to the
    shift that results in the maximal cross correlation.

    The module also calculates the polarization angle and the energy fluence
    from the fit parameters.
    """

    def __init__(self):
        self.__counter = 0
        self.begin()

    def begin(self):
        """
        begin method. This function is executed before the event loop.

        The antenna pattern provider is initialized here.
        """
        self.antenna_provider = antennapattern.AntennaPatternProvider()
        pass

    @register_run()
    def run(self, evt, station, det, debug=False, debug_plotpath=None,
            use_channels=[0, 1, 2, 3],
            bandpass=[100 * units.MHz, 500 * units.MHz],
            use_MC_direction=False):
        """
        run method. This function is executed for each event

        Parameters
        ---------
        evt
        station
        det
        debug: bool
            if True debug plotting is enables
        debug_plotpath: string or None
            if not None plots will be saved to a file rather then shown. Plots will
            be save into the `debug_plotpath` directory
        use_channels: array of ints
            the channel ids to use for the electric field reconstruction
            default: 0 - 3
        bandpass: [float, float]
            the lower and upper frequecy for which the analytic pulse is calculated.
            A butterworth filter of 10th order and a rectangular filter is applied.
            default 100 - 500 MHz
        use_MC_direction: bool
            use simulated direction instead of reconstructed direction
        """
        self.__counter += 1
        event_time = station.get_station_time()
        station_id = station.get_id()
        logger.info("event {}, station {}".format(evt.get_id(), station_id))
        if use_MC_direction and (station.get_sim_station() is not None):
            zenith = station.get_sim_station()[stnp.zenith]
            azimuth = station.get_sim_station()[stnp.azimuth]
            sim_present = True
        else:
            logger.warning("Using reconstructed angles as no simulation present")
            zenith = station[stnp.zenith]
            azimuth = station[stnp.azimuth]
            sim_present = False

        efield_antenna_factor, V, V_timedomain = get_array_of_channels(station, use_channels,
                                                                       det, zenith, azimuth, self.antenna_provider,
                                                                       time_domain=True)
        sampling_rate = station.get_channel(0).get_sampling_rate()
        n_samples_time = V_timedomain.shape[1]

        debug_obj = 0
        noise_RMS = det.get_noise_RMS(station.get_id(), 0)

        def obj_xcorr(params):
            if(len(params) == 3):
                slope, ratio2, phase2 = params
                ratio = (np.arctan(ratio2) + np.pi * 0.5) / np.pi  # project -inf..inf on 0..1
            elif(len(params) == 2):
                slope, ratio2 = params
                phase2 = 0
                ratio = (np.arctan(ratio2) + np.pi * 0.5) / np.pi  # project -inf..inf on 0..1
            elif(len(params) == 1):
                phase2 = 0
                ratio = 0
                slope = params[0]
            phase = np.arctan(phase2)  # project -inf..+inf to -0.5 pi..0.5 pi

            analytic_pulse_theta = pulse.get_analytic_pulse_freq(ratio, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass)
            analytic_pulse_phi = pulse.get_analytic_pulse_freq(1 - ratio, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass)
            chi2 = 0

            n_channels = len(V_timedomain)
            analytic_traces = np.zeros((n_channels, n_samples_time))
            positions = np.zeros(n_channels, dtype=np.int)
            max_xcorrs = np.zeros(n_channels)
            # first determine the position with the larges xcorr
            for iCh, trace in enumerate(V_timedomain):
                analytic_trace_fft = np.sum(efield_antenna_factor[iCh] * np.array([analytic_pulse_theta, analytic_pulse_phi]), axis=0)
                analytic_traces[iCh] = fft.freq2time(analytic_trace_fft, sampling_rate)
                xcorr = np.abs(hp.get_normalized_xcorr(trace, analytic_traces[iCh]))
                positions[iCh] = np.argmax(np.abs(xcorr)) + 1
                max_xcorrs[iCh] = xcorr.max()
                chi2 -= xcorr.max()
            logger.debug("ratio = {:.2f}, slope = {:.4g}, phase = {:.0f} ({:.4f}), chi2 = {:.4g}".format(ratio, slope, phase / units.deg, phase2, chi2))
            return chi2

        def obj_amplitude(params, slope, phase, pos, debug_obj=0):
            if(len(params) == 2):
                ampPhi, ampTheta = params
            elif(len(params) == 1):
                ampPhi = params[0]
                ampTheta = 0
            analytic_pulse_theta = pulse.get_analytic_pulse_freq(ampTheta, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass)
            analytic_pulse_phi = pulse.get_analytic_pulse_freq(ampPhi, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass)
            chi2 = 0

            if(debug_obj):
                fig, ax = plt.subplots(4, 2, sharex=True)

            n_channels = len(V_timedomain)
            analytic_traces = np.zeros((n_channels, n_samples_time))
            # first determine the position with the larges xcorr
            for iCh, trace in enumerate(V_timedomain):
                analytic_trace_fft = np.sum(efield_antenna_factor[iCh] * np.array([analytic_pulse_theta, analytic_pulse_phi]), axis=0)
                analytic_traces[iCh] = fft.freq2time(analytic_trace_fft, sampling_rate)

                argmax = np.argmax(np.abs(trace))
                imin = np.int(argmax - 30 * sampling_rate)
                imax = np.int(argmax + 50 * sampling_rate)

                tmp = np.sum(np.abs(trace[imin:imax] - np.roll(analytic_traces[iCh], pos)[imin:imax]) / noise_RMS)
                chi2 += tmp ** 2
                if(debug_obj):
                    ax[iCh][0].plot(trace, label='measurement')
                    ax[iCh][0].plot(np.roll(analytic_traces[iCh], pos), '--', label='fit')
                    ax[iCh][1].plot(trace - np.roll(analytic_traces[iCh], pos), label='delta')
                    ax[iCh][1].set_xlim(imin, imax)
            logger.debug("amp phi = {:.4g}, amp theta = {:.4g} , chi2 = {:.2g}".format(ampPhi, ampTheta, chi2))
            if(debug_obj):
                fig.suptitle("amp phi = {:.4g}, amp theta = {:.4g} , chi2 = {:.2g}".format(ampPhi, ampTheta, chi2))
                fig.tight_layout()
                plt.show()
            return chi2

        def obj_amplitude_slope(params, phase, pos, compare='hilbert', debug_obj=0):
            ampPhi, ampTheta, slope = params
            analytic_pulse_theta = pulse.get_analytic_pulse_freq(ampTheta, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass)
            analytic_pulse_phi = pulse.get_analytic_pulse_freq(ampPhi, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass)
            chi2 = 0
            if(debug_obj and self.i_slope_fit_iterations % 25 == 0):
                fig, ax = plt.subplots(4, 2, sharex=False, figsize=(20, 10))

            n_channels = len(V_timedomain)
            analytic_traces = np.zeros((n_channels, n_samples_time))
            # first determine the position with the larges xcorr
            channel_max = 0
            for trace in V_timedomain:
                if np.max(np.abs(trace)) > channel_max:
                    channel_max = np.max(np.abs(trace))
                    argmax = np.argmax(np.abs(trace))
                    imin = np.int(max(argmax - 50 * sampling_rate, 0))
                    imax = np.int(argmax + 50 * sampling_rate)
            for iCh, trace in enumerate(V_timedomain):
                analytic_trace_fft = np.sum(efield_antenna_factor[iCh] * np.array([analytic_pulse_theta, analytic_pulse_phi]), axis=0)
                analytic_traces[iCh] = fft.freq2time(analytic_trace_fft, sampling_rate)
                if compare == 'trace':
                    tmp = np.sum(np.abs(trace[imin:imax] - np.roll(analytic_traces[iCh], pos)[imin:imax]) ** 2) / noise_RMS ** 2
                elif compare == 'abs':
                    tmp = np.sum(np.abs(np.abs(trace[imin:imax]) - np.abs(np.roll(analytic_traces[iCh], pos)[imin:imax])) ** 2) / noise_RMS ** 2
                elif compare == 'hilbert':
                    tmp = np.sum(np.abs(np.abs(signal.hilbert(trace[imin:imax])) - np.abs(signal.hilbert(np.roll(analytic_traces[iCh], pos)[imin:imax]))) ** 2) / noise_RMS ** 2
                else:
                    raise NameError('Unsupported value for parameter "compare": {}. Value must be "trace", "abs" or "hilbert".'.format(compare))
                chi2 += tmp
                if(debug_obj and self.i_slope_fit_iterations % 25 == 0):
                    ax[iCh][1].plot(np.array(trace) / units.mV, label='measurement', color='blue')
                    ax[iCh][1].plot(np.roll(analytic_traces[iCh], pos) / units.mV, '--', label='fit', color='orange')
                    ax[iCh][1].plot(np.abs(signal.hilbert(trace)) / units.mV, linestyle=':', color='blue')
                    ax[iCh][1].plot(np.abs(signal.hilbert(np.roll(analytic_traces[iCh], pos))) / units.mV, ':', label='fit', color='orange')
                    # ax[iCh][1].plot(trace - np.roll(analytic_traces[iCh], pos), label='delta')
                    ax[iCh][0].plot(np.abs(fft.time2freq(trace, sampling_rate)) / units.mV, label='measurement')
                    ax[iCh][0].plot(np.abs(fft.time2freq(np.roll(analytic_traces[iCh], pos), sampling_rate)) / units.mV, '--', label='fit')
                    ax[iCh][0].set_xlim([0, 600])
                    ax[iCh][1].set_xlim([imin - 500, imax + 500])
                    ax[iCh][1].axvline(imin, linestyle='--', alpha=.8)
                    ax[iCh][1].axvline(imax, linestyle='--', alpha=.8)
            logger.debug("amp phi = {:.4g}, amp theta = {:.4g}, slope = {:.4g} chi2 = {:.8g}".format(ampPhi, ampTheta, slope, chi2))
            if(debug_obj and self.i_slope_fit_iterations % 25 == 0):
#                 ax[0][0].set_title("ratio = {:.2f} ({:.2f}), slope = {:.2g}, phase = {:.0f} ({:.2f}), chi2 = {:.2g}".format(ratio, ratio2, slope, phase / units.deg, phase2, chi2))
                fig.tight_layout()
                plt.show()
                self.i_slope_fit_iterations = 0
            self.i_slope_fit_iterations += 1
            return chi2

        def obj_amplitude_second_order(params, slope, phase, pos, compare='hilbert', debug_obj=0):
            ampPhi, ampTheta, second_order = params
            analytic_pulse_theta = pulse.get_analytic_pulse_freq(ampTheta, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass, quadratic_term=second_order, quadratic_term_offset=bandpass[0])
            analytic_pulse_phi = pulse.get_analytic_pulse_freq(ampPhi, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass, quadratic_term=second_order, quadratic_term_offset=bandpass[0])
            chi2 = 0
            if(debug_obj and self.i_slope_fit_iterations % 50 == 0):
                fig, ax = plt.subplots(5, 2, sharex=False, figsize=(20, 10))

            n_channels = len(V_timedomain)
            analytic_traces = np.zeros((n_channels, n_samples_time))
            # first determine the position with the larges xcorr
            channel_max = 0
            for trace in V_timedomain:
                if np.max(np.abs(trace)) > channel_max:
                    channel_max = np.max(np.abs(trace))
                    argmax = np.argmax(np.abs(trace))
                    imin = np.int(max(argmax - 50 * sampling_rate, 0))
                    imax = np.int(argmax + 50 * sampling_rate)
            for iCh, trace in enumerate(V_timedomain):
                analytic_trace_fft = np.sum(efield_antenna_factor[iCh] * np.array([analytic_pulse_theta, analytic_pulse_phi]), axis=0)
                analytic_traces[iCh] = fft.freq2time(analytic_trace_fft, sampling_rate)
                if compare == 'trace':
                    tmp = np.sum(np.abs(trace[imin:imax] - np.roll(analytic_traces[iCh], pos)[imin:imax]) ** 2) / noise_RMS ** 2
                elif compare == 'abs':
                    tmp = np.sum(np.abs(np.abs(trace[imin:imax]) - np.abs(np.roll(analytic_traces[iCh], pos)[imin:imax])) ** 2) / noise_RMS ** 2
                elif compare == 'hilbert':
                    tmp = np.sum(np.abs(np.abs(signal.hilbert(trace[imin:imax])) - np.abs(signal.hilbert(np.roll(analytic_traces[iCh], pos)[imin:imax]))) ** 2) / noise_RMS ** 2
                else:
                    raise NameError('Unsupported value for parameter "compare": {}. Value must be "trace", "abs" or "hilbert".'.format(compare))
                chi2 += tmp
                if(debug_obj and self.i_slope_fit_iterations % 50 == 0):
                    ax[iCh][1].plot(np.array(trace) / units.mV, label='measurement', color='blue')
                    ax[iCh][1].plot(np.roll(analytic_traces[iCh], pos) / units.mV, '--', label='fit', color='orange')
                    ax[iCh][1].plot(np.abs(signal.hilbert(trace)) / units.mV, linestyle=':', color='blue')
                    ax[iCh][1].plot(np.abs(signal.hilbert(np.roll(analytic_traces[iCh], pos))) / units.mV, ':', label='fit', color='orange')
                    # ax[iCh][1].plot(trace - np.roll(analytic_traces[iCh], pos), label='delta')
                    ax[iCh][0].plot(np.abs(fft.time2freq(trace, sampling_rate)) / units.mV, label='measurement')
                    ax[iCh][0].plot(np.abs(fft.time2freq(np.roll(analytic_traces[iCh], pos), sampling_rate)) / units.mV, '--', label='fit')
                    ax[iCh][0].set_xlim([0, 600])
                    ax[iCh][1].set_xlim([imin - 500, imax + 500])
                    ax[iCh][1].axvline(imin, linestyle='--', alpha=.8)
                    ax[iCh][1].axvline(imax, linestyle='--', alpha=.8)
            if(debug_obj and self.i_slope_fit_iterations % 50 == 0):
                sim_channel = station.get_sim_station().get_channel(0)[0]
                ax[4][0].plot(sim_channel.get_frequencies() / units.MHz, np.abs(pulse.get_analytic_pulse_freq(ampTheta, slope, phase, len(sim_channel.get_times()), sim_channel.get_sampling_rate(), bandpass=bandpass, quadratic_term=second_order)), '--', color='orange')
                ax[4][0].plot(sim_channel.get_frequencies() / units.MHz, np.abs(station.get_sim_station().get_channel(0)[0].get_frequency_spectrum()[1]), color='blue')
                ax[4][1].plot(sim_channel.get_frequencies() / units.MHz, np.abs(pulse.get_analytic_pulse_freq(ampPhi, slope, phase, len(sim_channel.get_times()), sim_channel.get_sampling_rate(), bandpass=bandpass, quadratic_term=second_order)), '--', color='orange')
                ax[4][1].plot(sim_channel.get_frequencies() / units.MHz, np.abs(sim_channel.get_frequency_spectrum()[2]), color='blue')
                ax[4][0].set_xlim([20, 500])
                ax[4][1].set_xlim([20, 500])
                # ax[4][0].set_yscale('log')
                # ax[4][1].set_yscale('log')
            logger.debug("amp phi = {:.4g}, amp theta = {:.4g}, slope = {:.4g} chi2 = {:.8g}".format(ampPhi, ampTheta, slope, chi2))
            if(debug_obj and self.i_slope_fit_iterations % 50 == 0):
        #                 ax[0][0].set_title("ratio = {:.2f} ({:.2f}), slope = {:.2g}, phase = {:.0f} ({:.2f}), chi2 = {:.2g}".format(ratio, ratio2, slope, phase / units.deg, phase2, chi2))
                fig.tight_layout()
                plt.show()
                self.i_slope_fit_iterations = 0
            self.i_slope_fit_iterations += 1
            return chi2

        method = "Nelder-Mead"
        # method = "BFGS"
        options = {'maxiter': 1000,
                   'disp': True}

        res = opt.minimize(obj_xcorr, x0=[-1], method=method, options=options)
        logger.info("slope xcorr fit, slope = {:.3g} with fmin = {:.3f}".format(res.x[0], res.fun))
        # plot objective function
        phase = 0
        ratio = 0
        slope = res.x[0]
        if slope > 0 or slope < -50:  # sanity check
            slope = -1.9
        analytic_pulse_theta_freq = pulse.get_analytic_pulse_freq(ratio, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass)
        analytic_pulse_phi_freq = pulse.get_analytic_pulse_freq(1 - ratio, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass)
        analytic_pulse_theta = pulse.get_analytic_pulse(ratio, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass)
        analytic_pulse_phi = pulse.get_analytic_pulse(1 - ratio, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass)

        n_channels = len(V_timedomain)
        analytic_traces = np.zeros((n_channels, n_samples_time))
        positions = np.zeros(n_channels, dtype=np.int)
        max_xcorrs = np.zeros(n_channels)
        for iCh, trace in enumerate(V_timedomain):
            analytic_trace_fft = np.sum(efield_antenna_factor[iCh] * np.array([analytic_pulse_theta_freq, analytic_pulse_phi_freq]), axis=0)
            analytic_traces[iCh] = fft.freq2time(analytic_trace_fft, sampling_rate)
            xcorr = np.abs(hp.get_normalized_xcorr(trace, analytic_traces[iCh]))
            positions[iCh] = np.argmax(np.abs(xcorr)) + 1
            max_xcorrs[iCh] = xcorr.max()
        pos = positions[np.argmax(max_xcorrs)]
        for iCh, trace in enumerate(V_timedomain):
            analytic_traces[iCh] = np.roll(analytic_traces[iCh], pos)

        res_amp = opt.minimize(obj_amplitude, x0=[1.], args=(slope, phase, pos, 0), method=method, options=options)
        logger.info("amplitude fit, Aphi = {:.3g} with fmin = {:.5e}".format(res_amp.x[0], res_amp.fun))
        Aphi = res_amp.x[0]
        Atheta = 0
        res_amp = opt.minimize(obj_amplitude, x0=[res_amp.x[0], 0], args=(slope, phase, pos, 0), method=method, options=options)
        logger.info("amplitude fit, Aphi = {:.3g} Atheta = {:.3g} with fmin = {:.5e}".format(res_amp.x[0], res_amp.x[1], res_amp.fun))
        Aphi = res_amp.x[0]
        Atheta = res_amp.x[1]
        # counts number of iterations in the slope fit. Used so we do not need to show the plots every iteration
        self.i_slope_fit_iterations = 0
        res_amp_slope = opt.minimize(obj_amplitude_slope, x0=[res_amp.x[0], res_amp.x[1], slope], args=(phase, pos, 'hilbert', False),
                                     method=method, options=options)

        # calculate uncertainties
        def Wrapper(params):
            return obj_amplitude_slope(params, phase, pos, 0)

        try:
            cov = covariance(Wrapper, res_amp_slope.x, 0.5, fast=True)
        except:
            cov = np.zeros((3, 3))
        logger.info("slope fit, Aphi = {:.3g}+-{:.3g} Atheta = {:.3g}+-{:.3g}, slope = {:.3g}+-{:.3g} with fmin = {:.5e}".format(res_amp_slope.x[0], cov[0, 0] ** 0.5,
                                                                                                                                     res_amp_slope.x[1], cov[1, 1] ** 0.5,
                                                                                                             res_amp_slope.x[2], cov[2, 2] ** 0.5, res_amp_slope.fun))
#         print(res_amp_slope)
        logger.info("covariance matrix \n{}".format(cov))
        if(cov[0, 0] > 0 and cov[1, 1] > 0 and cov[2, 2] > 0):
            logger.info("correlation matrix \n{}".format(hp.covariance_to_correlation(cov)))
        Aphi = res_amp_slope.x[0]
        Atheta = res_amp_slope.x[1]
        slope = res_amp_slope.x[2]
        Aphi_error = cov[0, 0] ** 0.5
        Atheta_error = cov[1, 1] ** 0.5
        slope_error = cov[2, 2] ** 0.5

         # plot objective function
        if 0:
            fo, ao = plt.subplots(1, 1)
            ss = np.linspace(-6, -0, 100)
            oos = [obj_amplitude_slope([res_amp_slope.x[0], res_amp_slope.x[1], s], phase, pos) for s in ss]
            ao.plot(ss, oos)

            n = 10
            x = np.linspace(res_amp_slope.x[0] * 0.6, res_amp_slope.x[0] * 1.4, n)
            y = np.linspace(-5, -1, n)
            X, Y = np.meshgrid(x, y)
            Z = np.zeros((n, n))
            for i in range(n):
                for j in range(n):
                    Z[i, j] = obj_amplitude_slope([X[i, j], X[i, j] * res_amp_slope.x[1] / res_amp_slope.x[0], Y[i, j]], phase, pos)

            fig, ax = plt.subplots(1, 1)
            ax.pcolor(X, Y, Z, cmap='viridis_r', vmin=res_amp_slope.fun, vmax=res_amp_slope.fun * 2)

        analytic_pulse_theta = pulse.get_analytic_pulse(Atheta, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass)
        analytic_pulse_phi = pulse.get_analytic_pulse(Aphi, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass)
        analytic_pulse_theta_freq = pulse.get_analytic_pulse_freq(Atheta, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass)
        analytic_pulse_phi_freq = pulse.get_analytic_pulse_freq(Aphi, slope, phase, n_samples_time, sampling_rate, bandpass=bandpass)

        analytic_pulse_theta = np.roll(analytic_pulse_theta, pos)
        analytic_pulse_phi = np.roll(analytic_pulse_phi, pos)
        station_trace = np.array([np.zeros_like(analytic_pulse_theta), analytic_pulse_theta, analytic_pulse_phi])

        electric_field = NuRadioReco.framework.electric_field.ElectricField(use_channels)
        electric_field.set_trace(station_trace, sampling_rate)
        energy_fluence = trace_utilities.get_electric_field_energy_fluence(electric_field.get_trace(), electric_field.get_times())
        electric_field.set_parameter(efp.signal_energy_fluence, energy_fluence)
        electric_field.set_parameter_error(efp.signal_energy_fluence, np.array([0, Atheta_error, Aphi_error]))
        electric_field.set_parameter(efp.cr_spectrum_slope, slope)
        electric_field.set_parameter(efp.zenith, zenith)
        electric_field.set_parameter(efp.azimuth, azimuth)
        # calculate high level parameters
        x = np.sign(Atheta) * np.abs(Atheta) ** 0.5
        y = np.sign(Aphi) * np.abs(Aphi) ** 0.5
        sx = Atheta_error * 0.5
        sy = Aphi_error * 0.5
        pol_angle = np.arctan2(abs(y), abs(x))
        pol_angle_error = 1. / (x ** 2 + y ** 2) * (y ** 2 * sx ** 2 + x ** 2 + sy ** 2) ** 0.5  # gaussian error propagation
        logger.info("polarization angle = {:.1f} +- {:.1f}".format(pol_angle / units.deg, pol_angle_error / units.deg))
        electric_field.set_parameter(efp.polarization_angle, pol_angle)
        electric_field.set_parameter_error(efp.polarization_angle, pol_angle_error)

        # compute expeted polarization
        site = det.get_site(station.get_id())
        exp_efield = hp.get_lorentzforce_vector(zenith, azimuth, hp.get_magnetic_field_vector(site))
        cs = coordinatesystems.cstrafo(zenith, azimuth, site=site)
        exp_efield_onsky = cs.transform_from_ground_to_onsky(exp_efield)
        exp_pol_angle = np.arctan2(exp_efield_onsky[2], exp_efield_onsky[1])
        logger.info("expected polarization angle = {:.1f}".format(exp_pol_angle / units.deg))
        electric_field.set_parameter(efp.polarization_angle_expectation, exp_pol_angle)
        res_amp_second_order = opt.minimize(obj_amplitude_second_order, x0=[res_amp_slope.x[0], res_amp_slope.x[1], 0], args=(slope, phase, pos, 'hilbert', False),
                                     method=method, options=options)
        second_order_correction = res_amp_second_order.x[2]
        electric_field.set_parameter(efp.cr_spectrum_quadratic_term, second_order_correction)
        # figure out the timing of the electric field
        voltages_from_efield = trace_utilities.get_channel_voltage_from_efield(station, electric_field, use_channels, det, zenith, azimuth, self.antenna_provider, False)
        correlation = np.zeros(voltages_from_efield.shape[1] + station.get_channel(use_channels[0]).get_trace().shape[0] - 1)
        channel_trace_start_times = []
        for channel_id in use_channels:
            channel_trace_start_times.append(station.get_channel(channel_id).get_trace_start_time())
        average_trace_start_time = np.average(channel_trace_start_times)
        for i_trace, v_trace in enumerate(voltages_from_efield):
            channel = station.get_channel(use_channels[i_trace])
            time_shift = geo_utl.get_time_delay_from_direction(zenith, azimuth, det.get_relative_position(station.get_id(), use_channels[i_trace])) - (channel.get_trace_start_time() - average_trace_start_time)
            voltage_trace = np.roll(np.copy(v_trace), int(time_shift * electric_field.get_sampling_rate()))
            correlation += signal.correlate(voltage_trace, channel.get_trace())
        toffset = (np.arange(0, correlation.shape[0]) - channel.get_trace().shape[0]) / electric_field.get_sampling_rate()
        electric_field.set_trace_start_time(-toffset[np.argmax(correlation)] + average_trace_start_time)
        station.add_electric_field(electric_field)

        if debug:
            analytic_traces = np.zeros((n_channels, n_samples_time))
            for iCh, trace in enumerate(V_timedomain):
                analytic_trace_fft = np.sum(efield_antenna_factor[iCh] * np.array([analytic_pulse_theta_freq, analytic_pulse_phi_freq]), axis=0)
                analytic_traces[iCh] = fft.freq2time(analytic_trace_fft, electric_field.get_sampling_rate())
                analytic_traces[iCh] = np.roll(analytic_traces[iCh], pos)
            fig, (ax2, ax2f) = plt.subplots(2, 1, figsize=(10, 8))
            lw = 2

            times = station.get_times() / units.ns
            ax2.plot(times, station.get_trace()[1] / units.mV * units.m, "-C0", label="analytic eTheta", lw=lw)
            ax2.plot(times, station.get_trace()[2] / units.mV * units.m, "-C1", label="analytic ePhi", lw=lw)
            tmax = times[np.argmax(station.get_trace()[2])]
            ax2.set_xlim(tmax - 40, tmax + 50)

            ff = station.get_frequencies() / units.MHz
            df = ff[1] - ff[0]
            ax2f.plot(ff[ff < 600], np.abs(station.get_frequency_spectrum()[1][ff < 600]) / df / units.mV * units.m, "-C0", label="analytic eTheta", lw=lw)
            ax2f.plot(ff[ff < 600], np.abs(station.get_frequency_spectrum()[2][ff < 600]) / df / units.mV * units.m, "-C1", label="analytic ePhi", lw=lw)

            if station.has_sim_station():
                sim_station = station.get_sim_station()
                logger.debug("station start time {:.1f}ns, relativ sim station time = {:.1f}".format(station.get_trace_start_time(), sim_station.get_trace_start_time()))
                df = (sim_station.get_frequencies()[1] - sim_station.get_frequencies()[0]) / units.MHz
                c = station.get_trace()[2].max() / sim_station.get_trace()[2].max()
                c = 1.
                ffsim = sim_station.get_frequencies()
                mask = (ffsim > 100 * units.MHz) & (ffsim < 500 * units.MHz)
                result = poly.polyfit(ffsim[mask], np.log10(np.abs(sim_station.get_frequency_spectrum()[2][mask]) / df / units.mV * units.m), 1, full=True)
                logger.info("polyfit result = {:.2g}  {:.2g}".format(*result[0]))
                ax2.plot(sim_station.get_times() / units.ns, sim_station.get_trace()[1] / units.mV * units.m * c, "--C0", label="simulation eTheta", lw=lw)
                ax2.plot(sim_station.get_times() / units.ns, sim_station.get_trace()[2] / units.mV * units.m * c, "--C1", label="simulation ePhi", lw=lw)
                ax2f.plot(sim_station.get_frequencies() / units.MHz, np.abs(sim_station.get_frequency_spectrum()[1]) / df / units.mV * units.m * c, "--C0", label="simulation eTheta", lw=lw)
                ax2f.plot(sim_station.get_frequencies() / units.MHz, np.abs(sim_station.get_frequency_spectrum()[2]) / df / units.mV * units.m * c, "--C1", label="simulation ePhi", lw=lw)

                ax2f.plot(ffsim / units.MHz, 10 ** (result[0][0] + result[0][1] * ffsim), "C3:")

            ax2.legend(fontsize="xx-small")
            ax2.set_xlabel("time [ns]")
            ax2.set_ylabel("electric-field [mV/m]")
            ax2f.set_ylim(1e-3, 5)
            ax2f.set_xlabel("Frequency [MHz]")
            ax2f.set_xlim(100, 500)
            ax2f.semilogy(True)
            if sim_present:
                sim = station.get_sim_station()
                fig.suptitle("Simulation: Zenith {:.1f}, Azimuth {:.1f}".format(np.rad2deg(sim[stnp.zenith]), np.rad2deg(sim[stnp.azimuth])))
            else:
                fig.suptitle("Data: reconstructed zenith {:.1f}, azimuth {:.1f}".format(np.rad2deg(zenith), np.rad2deg(azimuth)))
            fig.tight_layout()
            fig.subplots_adjust(top=0.95)
            if(debug_plotpath is not None):
                fig.savefig(os.path.join(debug_plotpath, 'run_{:05d}_event_{:06d}_efield.png'.format(evt.get_run_number(), evt.get_id())))
                plt.close(fig)

            # plot antenna response and channels
            fig, ax = plt.subplots(len(V), 3, sharex='col', sharey='col')
            for iCh in range(len(V)):
                mask = ff > 100
                ax[iCh, 0].plot(ff[mask], np.abs(efield_antenna_factor[iCh][0])[mask], label="theta, channel {}".format(use_channels[iCh]), lw=lw)
                ax[iCh, 0].plot(ff[mask], np.abs(efield_antenna_factor[iCh][1])[mask], label="phi, channel {}".format(use_channels[iCh]), lw=lw)
                ax[iCh, 0].legend(fontsize='xx-small')
                ax[iCh, 0].set_xlim(100, 500)
                ax[iCh, 1].set_xlim(400, 600)
                ax[iCh, 2].set_xlim(400, 600)
                ax[iCh, 1].plot(times, V_timedomain[iCh] / units.micro / units.V, lw=lw)
                ax[iCh, 1].plot(times, analytic_traces[iCh] / units.micro / units.V, '--', lw=lw)
                ax[iCh, 2].plot(times, (V_timedomain[iCh] - analytic_traces[iCh]) / units.micro / units.V, '-', lw=lw)
                ax[iCh, 0].set_ylabel("H [m]")
                ax[iCh, 1].set_ylabel(r"V [$\mu$V]")
                ax[iCh, 2].set_ylabel(r"$\Delta$V [$\mu$V]")
                RMS = det.get_noise_RMS(station.get_id(), 0)
                ax[iCh, 1].text(0.6, 0.8, 'S/N={:.1f}'.format(np.max(np.abs(V_timedomain[iCh])) / RMS), transform=ax[iCh, 1].transAxes)
            ax[0][2].set_ylim(ax[0][1].get_ylim())
            ax[-1, 1].set_xlabel("time [ns]")
            ax[-1, 2].set_xlabel("time [ns]")
            ax[-1, 0].set_xlabel("frequency [MHz]")
            fig.tight_layout()
            if(debug_plotpath is not None):
                fig.savefig(os.path.join(debug_plotpath, 'run_{:05d}_event_{:06d}_channels.png'.format(evt.get_run_number(), evt.get_id())))
                plt.close(fig)

    def end(self):
        pass



from NuRadioReco.modules.base.module import register_run
import h5py
import numpy as np
import NuRadioReco.framework.event
import NuRadioReco.framework.station
from NuRadioReco.modules.io.coreas import coreas

import logging
logger = logging.getLogger('readCoREASStation')


class readCoREASStation:

    def begin(self, input_files, station_id):
        """
        begin method

        initialize readCoREAS module

        Parameters
        ----------
        input_files: input files
            list of coreas hdf5 files
        station_id: station id
            id number of the radio station as defined in detector
        """
        self.__input_files = input_files
        self.__station_id = station_id
        self.__current_input_file = 0
        self.__current_event = 0

    @register_run()
    def run(self, detector):
        """
        Reads in all observers in the CoREAS files and returns a new simulated
        event for each.

        Parameters
        ----------
        detector: Detector object
            Detector description of the detector that shall be simulated

        """
        for input_file in self.__input_files:
            self.__current_event = 0
            with h5py.File(input_file, "r") as corsika:
                if list(corsika["highlevel"].values()) == []:
                    logger.warning(" No highlevel quantities in simulated hdf5 files, weights will be one")
                    weights = np.ones(len(corsika['CoREAS']['observers']))
                else:
                    weights = coreas.calculate_simulation_weights(list(corsika["highlevel"].values())[0]["antenna_position"])

                for i, (name, observer) in enumerate(corsika['CoREAS']['observers'].items()):
                    evt = NuRadioReco.framework.event.Event(self.__current_input_file, self.__current_event)  # create empty event
                    station = NuRadioReco.framework.station.Station(self.__station_id)
                    sim_station = coreas.make_sim_station(self.__station_id, corsika, observer,
                                 detector.get_channel_ids(self.__station_id), weights[i])
                    station.set_sim_station(sim_station)
                    sim_shower = coreas.make_sim_shower(corsika, observer, detector, self.__station_id)
                    evt.add_sim_shower(sim_shower)
                    evt.set_station(station)
                    self.__current_event += 1
                    yield evt
            self.__current_input_file += 1

    def end(self):
        pass



from __future__ import absolute_import, division, print_function, unicode_literals

def serialize(object):
    reply = {}
    for entry in object:
        reply[str(entry)] = object[entry]
    return reply

def deserialize(object, parameter_enum):
    reply = {}
    for entry in parameter_enum:
        if str(entry) in object:
            reply[entry] = object[str(entry)]
    return reply



import matplotlib.pyplot as plt
import numpy as np
import json

angles = np.linspace(-10.,10.,41)

angle_labels = ['{:.1f} degrees'.format(angle) for angle in angles]
chosen_angles = angle_labels[6:-6:4]
#chosen_angles = [angle_labels[18], angle_labels[33]]
#chosen_angles = [angle_labels[0], angle_labels[20], angle_labels[-1]]

filenames = ["snr_{:.1f}deg.json".format(-angle+35.) for angle in angles]

dict_list = []

for filename in filenames:
    with open(filename) as input:
        snr = json.load(input)
        dict_list.append(snr)

snr_halves = []

for angle_label, snr in zip(angle_labels, dict_list):

    SNRs = np.array( snr['SNRs'] )
    triggered = np.array(snr['triggered'])
    total_evts = float(snr['total_events'])
    efficiency = triggered/total_evts

    snr_half = SNRs[ efficiency > 0.5 ][0]
    snr_halves.append(snr_half)

    if angle_label in chosen_angles:
        plt.plot(SNRs, efficiency, label = angle_label)

plt.xlabel('SNR')
plt.xlim((0.5,5))
plt.ylim((0,1.01))
plt.legend()
plt.ylabel('Efficiency')
plt.savefig('snr_curves_more.pdf', format='pdf')
plt.show()

plt.plot(angles, snr_halves)
plt.xlabel(r'Vertex elevation angle [deg]')
plt.ylabel('SNR at 0.5 efficiency')
plt.ylim((1,3))
plt.xlim((-8,8))
plt.savefig('snr_half_eff.pdf', format='pdf')
plt.show()



from matplotlib import pyplot as plt
from NuRadioReco.detector import detector

det= detector.Detector()
fig, ax = plt.subplots(1, 1)
for site in ["A", "B", "C", "D", "E", "F", "G", "X", "Y", "Z"]:
    pos = det.get_absolute_position_site(site)
    ax.plot(pos[0], pos[1], 'o')
    ax.annotate(site, [pos[0], pos[1]])
    
ax.set_aspect("equal")
ax.set_xlabel("easting [m]")
ax.set_ylabel("northing [m]")
fig.tight_layout()
fig.savefig("map.png")
plt.show()

